/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*
 *
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

#include "mpiimpl.h"
#include "mpicomm.h"

/* -- Begin Profiling Symbol Block for routine MPI_Comm_create */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Comm_create = PMPI_Comm_create
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Comm_create  MPI_Comm_create
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Comm_create as PMPI_Comm_create
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_create
#define MPI_Comm_create PMPI_Comm_create

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_create


#if defined(USE_LID_LBI_CONTEXT) /* HK: */

int MPI_Comm_create(MPI_Comm comm, MPI_Group group, MPI_Comm *newcomm)
{
    static const char FCNAME[] = "MPI_Comm_create";
    int mpi_errno = MPI_SUCCESS;
    MPID_Comm *comm_ptr = NULL;
    MPID_Comm *newcomm_ptr;
    MPID_Group *group_ptr;

    MPIU_THREADPRIV_DECL;
    MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_CREATE);

    MPIR_ERRTEST_INITIALIZED_ORDIE();
    
    MPIU_THREAD_SINGLE_CS_ENTER("comm");
    MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_CREATE);
    MPIU_THREADPRIV_GET;

    /* Validate parameters, and convert MPI object handles to object pointers */
#   ifdef HAVE_ERROR_CHECKING
    {
        MPID_BEGIN_ERROR_CHECKS;
        {
	    MPIR_ERRTEST_COMM(comm, mpi_errno);
            if (mpi_errno) goto fn_fail;
	}
        MPID_END_ERROR_CHECKS;
	
	MPID_Comm_get_ptr( comm, comm_ptr );
	
        MPID_BEGIN_ERROR_CHECKS;
        {
            /* Validate comm_ptr */
            MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
	    /* If comm_ptr is not valid, it will be reset to null */

	    MPIR_ERRTEST_GROUP(group, mpi_errno);
            if (mpi_errno) goto fn_fail;
	}
        MPID_END_ERROR_CHECKS;
	
	MPID_Group_get_ptr( group, group_ptr );
    
        MPID_BEGIN_ERROR_CHECKS;
        {
	    /* Check the group ptr */
	    MPID_Group_valid_ptr( group_ptr, mpi_errno );
            if (mpi_errno) goto fn_fail;
        }
        MPID_END_ERROR_CHECKS;
    }
#   else
    {
	MPID_Comm_get_ptr( comm, comm_ptr );
	MPID_Group_get_ptr( group, group_ptr );
    }
#   endif

    FGP_DBG_PRINTF("group_ptr->size=%d, group_ptr->p_rank=%d, group_ptr->fgsize=%d, group_ptr->fgrank=%d, group_ptr->rtw_grp_map=%p\n", group_ptr->size, group_ptr->p_rank, group_ptr->fgsize, group_ptr->rank, group_ptr->rtw_grp_map); /* _p_rank_ */

    /* ... body of routine ...  */

    /* HK:
       Calling MPI_Comm_split here. The color is the world-id of the new group's fgrank=0
       and the key is the group's local fgrank.

       TODO: Optimization for small sized groups.

       In the case of USE_LID_LBI_CONTEXT algo, there is the question of how the
       group map is to be represented. Since the new comm_create algo relies on knowing who the
       group members are (in order for the local root to send/recv from them). The rtw_grp_map does
       not have the new_group_rank to outer_comm_rank mapping for that communication to take place.
       There are several ways of implementing this:
       1. One is to keep the local rank to outer_comm rank (rto) explicitly in the MPID_Group struct
          instead of the rtw map.
          Currently the rtw map representation for the map in MPID_Group MPID_Comm is the same. This
          uniformity has several advantages such as facilitating sharing of map between groups and
          communciators during MPI_Comm_create and MPI_Comm_group. It also helps avoid duplication
          of map among co-proclets as we point to a single instance of map within both MPID_Group
          and MPID_Comm structs. See the implementation of MPI_Comm_create below that uses MPICH2's
          context-id generation algo for details of sharing.
       2. Another option is to keep a pointer to the outer communicator and use its rtw_map (as a second
          level of indirection) to lookup the rank in outer comm. However, since communicators can be
          freed arbitrarily, this, however, will involve keeping reference counts to the
          outer comm's rtw map (See NOTE (A) below).
       3. Another option is to keep both rtw_grp_map and rto_grp_map (rto is the local rank to outer_comm
          explicit array). Use rto for send/recv of context-id information and then release that array and
          share rtw map as explained in point (1) above.
       4. Another option is to call MPI_Comm_split and use that algorithm already implemented. This is
          what is being done below. One advantage of comm_split over comm_create is that validity of group
          information is ensured (i.e., ensuring that groups do not overlap etc. due to a programmer error).
          
          
      One observation about the points 1,2 & 3 (above) is that the USE_LID_LBI_CONTEXT algo does not require
      all the members of the outer communciator to participate in the MPI_Comm_create operation; only those
      who are members of the new group. The MPICH2's context-id generation algo requires all in the outer
      communicator to participate and if some are not members of any group then they free the context-id
      by calling MPIR_Free_contextid().
       
    */
    
    
    if (comm_ptr->comm_kind == MPID_INTERCOMM) {

    }
    else { /* MPID_INTRACOMM */
        int color, key;
        if (group_ptr->rank != MPI_UNDEFINED) { /* HK: possibility of group_ptr->rank equal to MPI_UNDEFINED is coming from MPI_Group_incl. */
            color = RTWmapFindLeader(group_ptr->rtw_grp_map, group_ptr->fgsize);
            key = group_ptr->rank;                    }
        else{
            color = MPI_UNDEFINED;
            key = 0;
        }
        
        mpi_errno = MPI_Comm_split(comm, color, key, newcomm);
        if (mpi_errno) goto fn_fail;
        
        if (group_ptr->rank != MPI_UNDEFINED){ 
            
            assert (group_ptr->rtw_grp_map != NULL);

#   ifdef HAVE_ERROR_CHECKING
            {
                MPID_Comm_get_ptr( *newcomm, newcomm_ptr );
                
                MPID_BEGIN_ERROR_CHECKS;
                {
                    /* Validate newcomm_ptr */
                    MPID_Comm_valid_ptr( newcomm_ptr, mpi_errno );
                    /* If newcomm_ptr is not valid, it will be reset to null */                    
                    if (mpi_errno) goto fn_fail;
                }
                MPID_END_ERROR_CHECKS;
            }
#   else
            {
                MPID_Comm_get_ptr( *newcomm, newcomm_ptr );
            }
#   endif

            MPIR_Coshared_group_release(group_ptr);
            group_ptr->rtw_grp_map = newcomm_ptr->co_shared_vars->rtw_map; 
            group_ptr->ref_acrossCommGroup_countptr = newcomm_ptr->co_shared_vars->ref_acrossComm_countptr;
            MPIR_Comm_add_coshared_group_ref( group_ptr );
            /* Since the group has been provided, let the new communicator know
               about the group */
            newcomm_ptr->local_comm	    = 0;
            newcomm_ptr->local_group    = group_ptr;
            MPIR_Group_add_ref( group_ptr ); 
            newcomm_ptr->remote_group   = group_ptr;
            MPIR_Group_add_ref( group_ptr ); 
        }
        else{
            assert (*newcomm == MPI_COMM_NULL);
        }                    
    }


    /* ... end of body of routine ... */

  fn_exit:
    MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_CREATE);
    MPIU_THREAD_SINGLE_CS_EXIT("comm");
    return mpi_errno;

  fn_fail:
    /* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
    {
	mpi_errno = MPIR_Err_create_code(
	    mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**mpi_comm_create",
	    "**mpi_comm_create %C %G %p", comm, group, newcomm);
    }
#   endif
    mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
    goto fn_exit;
    /* --END ERROR HANDLING-- */
}






#else /* HK: #else USE_LID_LBI_CONTEXT is not defined, use MPICH2's context_id generation algorithm. */





/*@

MPI_Comm_create - Creates a new communicator

Input Parameters:
+ comm - communicator (handle) 
- group - group, which is a subset of the group of 'comm'  (handle) 

Output Parameter:
. comm_out - new communicator (handle) 

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_GROUP

.seealso: MPI_Comm_free
@*/
int MPI_Comm_create(MPI_Comm comm, MPI_Group group, MPI_Comm *newcomm)
{
    static const char FCNAME[] = "MPI_Comm_create";
    int mpi_errno = MPI_SUCCESS;
    MPID_Comm *comm_ptr = NULL;
    int i, j, n, *mapping = 0, *remote_mapping = 0, remote_size = -1, 
	new_context_id;
    MPID_Comm *newcomm_ptr;
    MPID_Group *group_ptr;
    MPIU_CHKLMEM_DECL(3);
    MPIU_THREADPRIV_DECL;
    MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_CREATE);

    MPIR_ERRTEST_INITIALIZED_ORDIE();
    
    MPIU_THREAD_SINGLE_CS_ENTER("comm");
    MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_CREATE);
    MPIU_THREADPRIV_GET;

    /* Validate parameters, and convert MPI object handles to object pointers */
#   ifdef HAVE_ERROR_CHECKING
    {
        MPID_BEGIN_ERROR_CHECKS;
        {
	    MPIR_ERRTEST_COMM(comm, mpi_errno);
            if (mpi_errno) goto fn_fail;
	}
        MPID_END_ERROR_CHECKS;
	
	MPID_Comm_get_ptr( comm, comm_ptr );
	
        MPID_BEGIN_ERROR_CHECKS;
        {
            /* Validate comm_ptr */
            MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
	    /* If comm_ptr is not valid, it will be reset to null */

	    MPIR_ERRTEST_GROUP(group, mpi_errno);
            if (mpi_errno) goto fn_fail;
	}
        MPID_END_ERROR_CHECKS;
	
	MPID_Group_get_ptr( group, group_ptr );
    
        MPID_BEGIN_ERROR_CHECKS;
        {
	    /* Check the group ptr */
	    MPID_Group_valid_ptr( group_ptr, mpi_errno );
            if (mpi_errno) goto fn_fail;
        }
        MPID_END_ERROR_CHECKS;
    }
#   else
    {
	MPID_Comm_get_ptr( comm, comm_ptr );
	MPID_Group_get_ptr( group, group_ptr );
    }
#   endif

    FGP_DBG_PRINTF("group_ptr->size=%d, group_ptr->p_rank=%d, group_ptr->fgsize=%d, group_ptr->fgrank=%d, group_ptr->rtw_grp_map=%p\n", group_ptr->size, group_ptr->p_rank, group_ptr->fgsize, group_ptr->fgrank, group_ptr->rtw_grp_map); /* _p_rank_ */
    
    /* ... body of routine ...  */
    
    /* Create a new communicator from the specified group members */

    /* If there is a context id cache in oldcomm, use it here.  Otherwise,
       use the appropriate algorithm to get a new context id. 
       Creating the context id is collective over the *input* communicator,
       so it must be created before we decide if this process is a 
       member of the group */
    /* In the multi-threaded case, MPIR_Get_contextid assumes that the
       calling routine already holds the single criticial section */
    if (comm_ptr->comm_kind == MPID_INTERCOMM) {
	if (!comm_ptr->local_comm) {
	    MPIR_Setup_intercomm_localcomm( comm_ptr );
	}
	new_context_id = MPIR_Get_contextid( comm_ptr->local_comm );
    }
    else {
        FGP_DBG_PRINTF("Before MPIR_Get_contextid ******************\n");
	new_context_id = MPIR_Get_contextid( comm_ptr ); 
        FGP_DBG_PRINTF("After MPIR_Get_contextid new_context_id=%d\n", new_context_id);
    }
    MPIU_ERR_CHKANDJUMP(new_context_id == 0, mpi_errno, MPI_ERR_OTHER, 
			"**toomanycomm" );


    /* Make sure that the processes for this group are contained within
       the input communicator.  Also identify the mapping from the ranks of 
       the old communicator to the new communicator.
       We do this by matching the lpids of the members of the group
       with the lpids of the members of the input communicator.
       It is an error if the group contains a reference to an lpid that 
       does not exist in the communicator.
       
       An important special case is groups (and communicators) that
       are subsets of MPI_COMM_WORLD.  This this case, the lpids are
       exactly the same as the ranks in comm world.  Currently, we
       don't take this into account, but if the code to handle the general 
       case is too messy, we'll add this in.
    */
    if (group_ptr->rank != MPI_UNDEFINED) { /* HK: possibility of group_ptr->rank equal to MPI_UNDEFINED is coming from MPI_Group_incl. */
        
	n = group_ptr->size; /* HK: This is the number of total HWP processes. Note: n is being used later, so must be properly initialized */
        
#if !defined(FINEGRAIN_MPI) /*HK: We don't need to fill in the mapping array, since the information is
                              already available in the group rtw map. Note: group is a subset of comm
                              based on which the newcomm is created */
        MPID_VCR *vcr;
	int      vcr_size;

	MPIU_CHKLMEM_MALLOC(mapping,int*,n*sizeof(int),mpi_errno,"mapping"); 
	if (comm_ptr->comm_kind == MPID_INTERCOMM) {
	    vcr      = comm_ptr->local_vcr;
	    vcr_size = comm_ptr->local_size;
	}
	else {
	    vcr      = comm_ptr->vcr;
	    vcr_size = comm_ptr->remote_size; 
	}
	
	for (i=0; i<n; i++) {
	    /* Mapping[i] is the rank in the communicator of the process that
	       is the ith element of the group */
	    /* We use the appropriate vcr, depending on whether this is
	       an intercomm (use the local vcr) or an intracomm (remote vcr) 
	       Note that this is really the local mapping for intercomm
	       and remote mapping for the intracomm */
	    /* FIXME : BUBBLE SORT */
	    /* FIXME : NEEDS COMM_WORLD SPECIALIZATION */
	    mapping[i] = -1;
	    for (j=0; j<vcr_size; j++) {
		int comm_lpid;
		MPID_VCR_Get_lpid( vcr[j], &comm_lpid );
		if (comm_lpid == group_ptr->lrank_to_lpid[i].lpid) {
		    mapping[i] = j; /* HK: This is mapping to the HWP rank */
		    break;
		}
	    }
	    MPIU_ERR_CHKANDJUMP1(mapping[i] == -1,mpi_errno,
				 MPI_ERR_GROUP,
				 "**groupnotincomm", "**groupnotincomm %d", i );
	}
	if (comm_ptr->comm_kind == MPID_INTRACOMM) {
	    /* If this is an intra comm, we've really determined the
	       remote mapping */
	    remote_mapping = mapping;
	    remote_size    = n;
	    mapping        = 0;
	}
#endif /* HK: end of #if !defined(FINEGRAIN_MPI) */
        
	/* Get the new communicator structure and context id */
	
	mpi_errno = MPIR_Comm_create( &newcomm_ptr );
	if (mpi_errno) goto fn_fail;

	newcomm_ptr->context_id	    = new_context_id;
	newcomm_ptr->rank	    = group_ptr->rank; 
#if defined(FINEGRAIN_MPI)
        MPIR_Intra_newcomm_init(newcomm_ptr, group_ptr->p_rank, group_ptr->fgsize); /* _p_rank_ */
#endif
	newcomm_ptr->comm_kind	    = comm_ptr->comm_kind;
	/* Since the group has been provided, let the new communicator know
	   about the group */
        newcomm_ptr->local_comm	    = 0;
	newcomm_ptr->local_group    = group_ptr;
	MPIR_Group_add_ref( group_ptr ); 

	if (comm_ptr->comm_kind == MPID_INTRACOMM) {
	    newcomm_ptr->remote_group   = group_ptr;
	    MPIR_Group_add_ref( group_ptr ); 
	    newcomm_ptr->recvcontext_id = new_context_id;
	    newcomm_ptr->remote_size    = newcomm_ptr->local_size = n; /* HK: Double-check This is the total number of HWP processes. */
	}
	else { /* comm_ptr->comm_kind == MPID_INTERCOMM */
	    int rinfo[2];
	    newcomm_ptr->local_size   = group_ptr->size; 
	    newcomm_ptr->remote_group = 0;
	    /* There is an additional step.  We must communicate the information
	       on the local context id and the group members, given by the ranks
	       so that the remote process can construct the appropriate VCRT 
	       First we exchange group sizes and context ids.  Then the 
	       ranks in the remote group, from which the remote VCRT can 
	       be constructed.  We can't use NMPI_Sendrecv since we need to 
	       use the "collective" context in the original intercommunicator */
	    if (comm_ptr->rank == 0) {
		int info[2];
		info[0] = new_context_id;
		info[1] = group_ptr->size;

		mpi_errno = MPIC_Sendrecv(info, 2, MPI_INT, 
					  0, 0, 
					  rinfo, 2, MPI_INT,
					  0, 0, comm, MPI_STATUS_IGNORE );
		if (mpi_errno) { MPIU_ERR_POP( mpi_errno ); }
		newcomm_ptr->recvcontext_id = rinfo[0];
		remote_size                 = rinfo[1];
		    
		MPIU_CHKLMEM_MALLOC(remote_mapping,int*,
				    remote_size*sizeof(int),
				    mpi_errno,"remote_mapping");

		/* Populate and exchange the ranks */
		mpi_errno = MPIC_Sendrecv( mapping, group_ptr->size, MPI_INT, 
					   0, 0, 
					   remote_mapping, remote_size, MPI_INT,
					   0, 0, comm, MPI_STATUS_IGNORE );
		if (mpi_errno) { MPIU_ERR_POP( mpi_errno ); }

		/* Broadcast to the other members of the local group */
                MPIR_Nest_incr();
		NMPI_Bcast( rinfo, 2, MPI_INT, 0, 
			    comm_ptr->local_comm->handle );
		NMPI_Bcast( remote_mapping, remote_size, MPI_INT, 0, 
			    comm_ptr->local_comm->handle );
                MPIR_Nest_decr();
	    }
	    else {
		/* The other processes */
		/* Broadcast to the other members of the local group */
                MPIR_Nest_incr();
		NMPI_Bcast( rinfo, 2, MPI_INT, 0, 
			    comm_ptr->local_comm->handle );
		newcomm_ptr->recvcontext_id = rinfo[0];
		remote_size                 = rinfo[1];
		MPIU_CHKLMEM_MALLOC(remote_mapping,int*,
				    remote_size*sizeof(int),
				    mpi_errno,"remote_mapping");
		NMPI_Bcast( remote_mapping, remote_size, MPI_INT, 0, 
			    comm_ptr->local_comm->handle );
                MPIR_Nest_decr();
	    }
	    newcomm_ptr->remote_size    = remote_size;
	    /* Now, everyone has the remote_mapping, and can apply that to 
	       the vcr table.  That step is done below, as it is the 
	       same step as needed for the intracommunicator */

	    /* Setup the communicator's local vc table from the local 
	       mapping. */
	    n = group_ptr->size;
	    MPID_VCRT_Create( n, &newcomm_ptr->local_vcrt ); 
	    MPID_VCRT_Get_ptr( newcomm_ptr->local_vcrt, 
			       &newcomm_ptr->local_vcr );
	    for (i=0; i<n; i++) {
		/* For rank i in the new communicator, find the corresponding
		   rank in the input communicator */
		MPID_VCR_Dup( comm_ptr->local_vcr[mapping[i]], 
			      &newcomm_ptr->local_vcr[i] );
		
	    }
	}  /* HK: closing brace of else{ comm_ptr->comm_kind == MPID_INTERCOMM */

 	/* Setup the communicator's vc table.  This is for the remote group,
	   which is the same as the local group for intracommunicators */


        /* HK: pointing to vcrt_world and vcr_world in MPIR_Intra_newcomm_init() above.
           instead of creating and then doing
           MPID_VCR_Dup( comm_ptr->vcr[i], &newcomm_ptr->vcr[i] ); in a loop. 
        */
#if !defined(FINEGRAIN_MPI)
	n = remote_size;
	MPID_VCRT_Create( n, &newcomm_ptr->vcrt );
	MPID_VCRT_Get_ptr( newcomm_ptr->vcrt, &newcomm_ptr->vcr );
	for (i=0; i<n; i++) {
	    /* For rank i in the new communicator, find the corresponding
	       rank in the input communicator */
	    MPID_VCR_Dup( comm_ptr->vcr[remote_mapping[i]], 
			  &newcomm_ptr->vcr[i] );
	}
#endif

#if defined(FINEGRAIN_MPI)        
        /* HK: Looking up pointer to shared rtw map in contextLeader_hshtbl.
           If it does not exist (i.e. first proclet in MPI_Comm_create()), then create it. */
        int LeaderWorldRank = RTWmapFindLeader(group_ptr->rtw_grp_map, newcomm_ptr->totprocs);
        cLitemptr stored = NULL;
        CL_LookupHashFind(contextLeader_hshtbl, newcomm_ptr->context_id, LeaderWorldRank, &stored);
        if(stored){
            MPIR_Comm_add_coshared_all_ref(stored->coproclet_shared_vars);
            newcomm_ptr->co_shared_vars = ((Coproclet_shared_vars_t *)(stored->coproclet_shared_vars));
            MPIR_Copy_ch3i_comm_ch(newcomm_ptr, ((Coproclet_shared_vars_t *)(stored->coproclet_shared_vars))->ch3i_ptr);
            FGP_DBG_PRINTF("Already present in contextLeader_hshtbl, context-id=%d, LeaderWorldRank=%d, newcomm_ptr->co_shared_vars->rtw_map=%p\n", newcomm_ptr->context_id, LeaderWorldRank, newcomm_ptr->co_shared_vars->rtw_map);
            /* HK: Freeing the group rtw map and making it point to the shared rtw map for that group.
               This is being done to avoid having a copy of the rtw map in all co-proclets of that group.*/
            MPIR_Coshared_group_release(group_ptr);
            group_ptr->rtw_grp_map = newcomm_ptr->co_shared_vars->rtw_map;
            group_ptr->ref_acrossCommGroup_countptr = newcomm_ptr->co_shared_vars->ref_acrossComm_countptr;
            MPIR_Comm_add_coshared_group_ref( group_ptr );
        }
        else {
            /*
              HK:
              For the release of the group map, we use reference counting. The reference
              count will be incremented by every proclet sharing that map.
              It will be decremented in MPI_Comm_free and the map will only be released when the
              reference count goes to zero. See comment in MPID_Group struct.
              NOTE (A): Note that this reference counting is separate
              from the communicator's current reference counting and will need to be stored in
              context-id hashtable so that each proclet can access it and update it. Remember that
              each proclet has its own communicator instance, while the group map is shared.
              Hence the reference counting for release of group map has to be associated with
              it through the hashtable. Another thing to remember is that comm_free should be
              collective operation. MPICH2 has a "FIXME" comment associated with it saying that
              it should be collective but it is not yet done that way.
              
              In the following, since the rtw map is not being released in MPI_Group_free(), it is not necessary for
              the first proclet to duplicate group_ptr->rtw_grp_map and later release it through
              RTWmapKill. Simply assigning newcomm_ptr->co_shared_vars->rtw_map to group_ptr->rtw_grp_map for first proclet.
              // newcomm_ptr->co_shared_vars->rtw_map = RTWmapDuplicate(group_ptr->rtw_grp_map, newcomm_ptr->totprocs);
            */
            newcomm_ptr->co_shared_vars = (Coproclet_shared_vars_t *)MPIU_Malloc(sizeof(Coproclet_shared_vars_t));
            /* --BEGIN ERROR HANDLING-- */
            if ( (newcomm_ptr->co_shared_vars == NULL) ) {                    
                printf("%s:Line_%d my_fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank);
                mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
                goto fn_fail;
            }
            /* --END ERROR HANDLING-- */            

            newcomm_ptr->co_shared_vars->rtw_map = group_ptr->rtw_grp_map;
            newcomm_ptr->co_shared_vars->co_barrier_vars =  (struct coproclet_barrier_vars *)MPIU_Malloc(sizeof(struct coproclet_barrier_vars));
            /* --BEGIN ERROR HANDLING-- */
            if ( (newcomm_ptr->co_shared_vars->co_barrier_vars == NULL) ) {                    
                printf("%s:Line_%d my_fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank);
                mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
                goto fn_fail;
            }
            /* --END ERROR HANDLING-- */            

            newcomm_ptr->co_shared_vars->co_barrier_vars->coproclet_signal = 0;
            newcomm_ptr->co_shared_vars->co_barrier_vars->coproclet_counter = 0;
            newcomm_ptr->co_shared_vars->co_barrier_vars->leader_signal = 0;
            /* Notify the device of this new communicator */
            MPID_Dev_comm_create_hook( newcomm_ptr ); /* HK: This macro resolves to MPIDI_CH3I_comm_create( ) */

            if ( group_ptr->ref_acrossCommGroup_countptr ) {
                MPIR_Comm_init_coshared_withinComm_ref(newcomm_ptr->co_shared_vars);
                newcomm_ptr->co_shared_vars->ref_acrossComm_countptr = group_ptr->ref_acrossCommGroup_countptr;
                MPIR_Comm_add_coshared_acrossComm_ref(newcomm_ptr->co_shared_vars);
            } else {
                MPIR_Comm_init_coshared_all_ref(newcomm_ptr->co_shared_vars);
                group_ptr->ref_acrossCommGroup_countptr = newcomm_ptr->co_shared_vars->ref_acrossComm_countptr;
                MPIR_Comm_add_coshared_group_ref(group_ptr);
            }
            newcomm_ptr->co_shared_vars->ch3i_ptr = &(newcomm_ptr->ch);
            stored = NULL;
            CL_LookupHashInsert(contextLeader_hshtbl, newcomm_ptr->context_id, LeaderWorldRank, newcomm_ptr->co_shared_vars, &stored);
            if(!stored){
                printf("Store failure in contextLeader_hshtbl\n");
                MPID_Abort(NULL, MPI_SUCCESS, -1, NULL); 
                exit(-1);
            }
            FGP_DBG_PRINTF("Inserting in contextLeader_hshtbl, context-id=%d, LeaderWorldRank=%d, newcomm_ptr->co_shared_vars->rtw_map=%p\n", newcomm_ptr->context_id, LeaderWorldRank, newcomm_ptr->co_shared_vars->rtw_map);
            /* HK: TODO IMPORTANT. When should the shared rtw map be free'd? Should I have ref_counts
               for the shared rtw map, which is incremented for every proclet that is pointing to that
               shared rtw map and the last proclet to call MPI_Comm_free decrements it to zero and frees it?
            */
        }


#else        
	/* Notify the device of this new communicator */
	MPID_Dev_comm_create_hook( newcomm_ptr ); /* HK: This macro resolves to MPIDI_CH3I_comm_create( ) */
#endif

	*newcomm = newcomm_ptr->handle;
    }
    else {
	/* This process is not in the group */
	MPIR_Free_contextid( new_context_id );
	*newcomm = MPI_COMM_NULL;

	/* Dummy to complete collective ops in the intercomm case */
	if (comm_ptr->comm_kind == MPID_INTERCOMM) {
	    int rinfo[2], ic_remote_size, *ic_remote_mapping = 0;
            MPIR_Nest_incr();
	    NMPI_Bcast( rinfo, 2, MPI_INT, 0, 
			comm_ptr->local_comm->handle );
	    ic_remote_size = rinfo[1];
	    MPIU_CHKLMEM_MALLOC(ic_remote_mapping,int*,
				ic_remote_size*sizeof(int),
				mpi_errno,"ic_remote_mapping");
	    NMPI_Bcast( ic_remote_mapping, ic_remote_size, MPI_INT, 0, 
			comm_ptr->local_comm->handle );
            MPIR_Nest_decr();
	}
    }
    
    /* ... end of body of routine ... */

    /* mpi_errno = MPID_Comm_create(); */
    if (mpi_errno != MPI_SUCCESS) goto fn_fail;

  fn_exit:
    MPIU_CHKLMEM_FREEALL(); /* HK: mapping[] is being freed here. */
    MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_CREATE);
    MPIU_THREAD_SINGLE_CS_EXIT("comm");
    return mpi_errno;

  fn_fail:
    /* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
    {
	mpi_errno = MPIR_Err_create_code(
	    mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**mpi_comm_create",
	    "**mpi_comm_create %C %G %p", comm, group, newcomm);
    }
#   endif
    mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
    goto fn_exit;
    /* --END ERROR HANDLING-- */
}


#endif /*(USE_LID_LBI_CONTEXT) */


#if defined(USE_LID_LBI_CONTEXT)

#define NEED_SORTING 0
#define MSGS_TO_NEIGHBOR 3
enum rtwmap_flag_t {CREATE_RTWMAP, SHARE_RTWMAP, CREATE_OR_SHARE_RTWMAP};

typedef struct NeighborLocalrank{
    int outercomm_rank;
    int new_localrank;
} NeighborLocalrank_t;

int qsortNeighborsArray(const void* neighbors1_ptr, const void* neighbors2_ptr){
    int c1 = ((NeighborLocalrank_t *)(neighbors1_ptr))->outercomm_rank;
    int c2 = ((NeighborLocalrank_t *)(neighbors2_ptr))->outercomm_rank;
    if(c1 < c2) return (-1);
    else if(c1 == c2) return (0);
    else return (1);    
}


inline void MPIU_Sort_neighbors( NeighborLocalrank_t *neighbors, int size )
{
    qsort(neighbors, size, sizeof(NeighborLocalrank_t), qsortNeighborsArray);
}


#undef FUNCNAME
#define FUNCNAME MPI_Populate_comm

int MPIR_Populate_comm( MPID_Comm *outercomm_ptr, MPID_Comm *newcomm_ptr, int *group_array,
                        int HWP_size, int new_CID, int new_localrank, int new_totprocs, enum rtwmap_flag_t rtw_flag )
{
    static const char FCNAME[] = "MPIR_Populate_comm";
    int mpi_errno = MPI_SUCCESS;
    int i;
    newcomm_ptr->context_id	    = new_CID; 
    newcomm_ptr->local_size	    = HWP_size; /* HK: This is the total number of HWP processes */
    newcomm_ptr->comm_kind	    = outercomm_ptr->comm_kind;
    
    newcomm_ptr->recvcontext_id = newcomm_ptr->context_id;
    newcomm_ptr->remote_size = HWP_size; /* HK: TODO Double-check. This is the total number of HWP processes */
    newcomm_ptr->rank = new_localrank; 
    MPIR_Intra_newcomm_init(newcomm_ptr, outercomm_ptr->p_rank, new_totprocs); /* _p_rank_ */


    /* HK: Create or share rtw map in cidLookup_hshtbl.
     */
    cidLookupHashItemptr stored = NULL;
    CidLookupHashFind(cidLookup_hshtbl, newcomm_ptr->context_id, &stored);
    if ( (rtw_flag == SHARE_RTWMAP) ||
         ((rtw_flag == CREATE_OR_SHARE_RTWMAP) && (stored != NULL)) ) { /* HK: Share the rtwmap */
        assert(stored != NULL);
        MPIR_Comm_add_coshared_all_ref(stored->coproclet_shared_vars);
        newcomm_ptr->co_shared_vars = ((Coproclet_shared_vars_t *)(stored->coproclet_shared_vars));
        MPIR_Copy_ch3i_comm_ch(newcomm_ptr, ((Coproclet_shared_vars_t *)(stored->coproclet_shared_vars))->ch3i_ptr);
        FGP_DBG_PRINTF("Already present in cidLookup_hshtbl, context-id=%d, newcomm_ptr->co_shared_vars->rtw_map=%p\n", newcomm_ptr->context_id, newcomm_ptr->co_shared_vars->rtw_map);        
    }
    else if ( (rtw_flag == CREATE_RTWMAP) ||
              ((rtw_flag == CREATE_OR_SHARE_RTWMAP) && (stored == NULL)) ) { /* HK: Create the rtwmap */
        assert(stored == NULL);          
        newcomm_ptr->co_shared_vars = (Coproclet_shared_vars_t *)MPIU_Malloc(sizeof(Coproclet_shared_vars_t));
        /* --BEGIN ERROR HANDLING-- */
        if ((newcomm_ptr->co_shared_vars == NULL)) {                    
            printf("%s:Line_%d my_fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank );
            mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
            goto fn_fail;
        }
        /* --END ERROR HANDLING-- */
        
        newcomm_ptr->co_shared_vars->rtw_map = (RTWmap *)RTWmapCreate(new_totprocs);
        assert( newcomm_ptr->co_shared_vars->rtw_map != NULL );
        int *rtw_blockinsert = (int*) MPIU_Malloc(sizeof(int) * new_totprocs);
        assert(rtw_blockinsert != 0);
        for (i=0; i<newcomm_ptr->totprocs; i++) {
            /* Initializing worldrank to -1 */
            int worldrank = -1, key;
            key = group_array[i];
            RTWPmapFind(outercomm_ptr->co_shared_vars->rtw_map, key, &worldrank, NULL);
            key = i; /* HK: This is the local fgrank relative to this new group */
            /* HK: Store item in newcomm_ptr rtw map which has i as
               the local fgrank relative to the new communicator.
               The corresponding worldrank in outer communicator is
               determined and saved against fgrank=i in the newcomm_ptr rtw map. 
            */
            rtw_blockinsert[key] = worldrank;                    
        }
        
        RTWmapBlockInsert(newcomm_ptr->co_shared_vars->rtw_map, new_totprocs, rtw_blockinsert);
        MPIU_Free(rtw_blockinsert);
        
        newcomm_ptr->co_shared_vars->co_barrier_vars =  (struct coproclet_barrier_vars *)MPIU_Malloc(sizeof(struct coproclet_barrier_vars));
        
        /* --BEGIN ERROR HANDLING-- */
        if ((newcomm_ptr->co_shared_vars->co_barrier_vars == NULL)) {                    
            printf("%s:Line_%d my_fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank);
            mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
            goto fn_fail;
        }
        /* --END ERROR HANDLING-- */
        
        newcomm_ptr->co_shared_vars->co_barrier_vars->coproclet_signal = 0;
        newcomm_ptr->co_shared_vars->co_barrier_vars->coproclet_counter = 0;
        newcomm_ptr->co_shared_vars->co_barrier_vars->leader_signal = 0;
        /* Notify the device of this new communicator */
        MPID_Dev_comm_create_hook( newcomm_ptr );
        
        MPIR_Comm_init_coshared_all_ref(newcomm_ptr->co_shared_vars);
        newcomm_ptr->co_shared_vars->ch3i_ptr = &(newcomm_ptr->ch);
        stored = NULL;
        CidLookupHashInsert(cidLookup_hshtbl, newcomm_ptr->context_id, newcomm_ptr->co_shared_vars, &stored);
        if(!stored){
            printf("Store failure in contextLeader_hshtbl\n");
            MPIU_ERR_CHKANDJUMP(stored == 0, mpi_errno, MPI_ERR_OTHER, 
                                "**hshstorefail" );
            exit(-1);
        }
        FGP_DBG_PRINTF("Inserting in cidLookup_hshtbl, context-id=%d, newcomm_ptr->co_shared_vars->rtw_map=%p\n", newcomm_ptr->context_id, newcomm_ptr->co_shared_vars->rtw_map);
    }
    else {
        MPIU_Assert(0); /* Should not be reached */
    }
        
    /* Inherit the error handler (if any) */ 
    newcomm_ptr->errhandler = outercomm_ptr->errhandler;
    if (outercomm_ptr->errhandler) {
        MPIR_Errhandler_add_ref( outercomm_ptr->errhandler );
    }
    
 fn_exit:
    return mpi_errno;
    
 fn_fail:
    /* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
    {
	mpi_errno = MPIR_Err_create_code(
                                         mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**mpi_populate_comm",
                                         "**mpi_populate_comm %p", newcomm_ptr);
    }
#   endif
    mpi_errno = MPIR_Err_return_comm( outercomm_ptr, FCNAME, mpi_errno );
    goto fn_exit;
    /* --END ERROR HANDLING-- */
    
}


/* -- Begin Profiling Symbol Block for routine MPIX_Comms_create */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPIX_Comms_create = PMPIX_Comms_create
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPIX_Comms_create  MPIX_Comms_create
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPIX_Comms_create as PMPIX_Comms_create
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPIX_Comms_create
#define MPIX_Comms_create PMPIX_Comms_create

#endif

#undef FUNCNAME
#define FUNCNAME MPIX_Comms_create

/* HK: neighbors_array includes myself in specified rank order
   and num_neighbors includes self too.
 */

int MPIX_Comms_create(MPI_Comm comm, int *neighbors_array,
                     int num_neighbors, MPI_Comm *neighborscomms)
{
    static const char FCNAME[] = "MPIX_Comms_create";
    int mpi_errno = MPI_SUCCESS;
    MPID_Comm *comm_ptr = NULL;
    MPID_Comm **neighborscomm_ptr = NULL;
    int my_neighbors_CID = -1;
    int my_neighbors_LBI_bit = -1;
    int i, size = -1, fgrank = -1;
    int my_neighborscomm_rank = -1;
    int neighbor_rank, neighbor_newLocalRank;
    NeighborLocalrank_t * sorted_neighbors = NULL;
    int sorted = TRUE;
    int prev_rank = -1;
        
    MPIU_CHKLMEM_DECL(2);
    MPID_MPI_STATE_DECL(MPID_STATE_MPIX_COMMS_CREATE);

    MPIR_ERRTEST_INITIALIZED_ORDIE();
    
    MPID_MPI_FUNC_ENTER(MPID_STATE_MPIX_COMMS_CREATE);

    /* Validate parameters, and convert MPI object handles to object pointers */
#   ifdef HAVE_ERROR_CHECKING
    {
        MPID_BEGIN_ERROR_CHECKS;
        {
	    MPIR_ERRTEST_COMM(comm, mpi_errno);
            if (mpi_errno) goto fn_fail;
	}
        MPID_END_ERROR_CHECKS;
	
	MPID_Comm_get_ptr( comm, comm_ptr );
	
        MPID_BEGIN_ERROR_CHECKS;
        {
            /* Validate comm_ptr */
            MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
	    /* If comm_ptr is not valid, it will be reset to null */

            if (mpi_errno) goto fn_fail;
	}
        MPID_END_ERROR_CHECKS;
	
    }
#   else
    {
	MPID_Comm_get_ptr( comm, comm_ptr );
    }
#   endif


    /* ... body of routine ...  */

    size        = comm_ptr->local_size;
    fgrank      = comm_ptr->rank;

    MPIU_CHKLMEM_MALLOC(neighborscomm_ptr, MPID_Comm **, sizeof(MPID_Comm *)*(num_neighbors),mpi_errno, "neighborscomm_ptr");
    if ( NEED_SORTING ){
        MPIU_CHKLMEM_MALLOC(sorted_neighbors, NeighborLocalrank_t *, sizeof(NeighborLocalrank_t)*(num_neighbors),mpi_errno, "sorted_neighbors");
    }

    prev_rank = neighbors_array[0];
    for (i=0; i<num_neighbors; i++) {
        if ( NEED_SORTING ){
            sorted_neighbors[i].outercomm_rank = neighbors_array[i];
            sorted_neighbors[i].new_localrank = i;
            
            if ( prev_rank > neighbors_array[i] ){
                sorted = FALSE;
            }
            prev_rank = neighbors_array[i];
            
            if( fgrank == neighbors_array[i]){
                my_neighborscomm_rank = i;
            }
        } else {
            if( fgrank == neighbors_array[i]){
                my_neighborscomm_rank = i;
                break;
            }            
        }
    }
    assert ( my_neighborscomm_rank >=0 );

    if (FALSE == sorted && NEED_SORTING){
        MPIU_Sort_neighbors( sorted_neighbors, num_neighbors );
    }
    
    if ( initialize_LBI_mask ){
        MPIR_Init_LBI();
    }
    assert (LBI_mask != NULL);


    /* Update my LBI mask */    
    my_neighbors_LBI_bit = MPIR_Find_LBI_bit( LBI_mask );
    assert( (my_neighbors_LBI_bit >=0) &&
            (my_neighbors_LBI_bit < (MPIR_CONTEXT_INT_BITS*MAX_LBI_MASK)) ); 
    my_neighbors_CID = MPIR_Get_CID(my_fgrank, my_neighbors_LBI_bit); /* passing my_fgrank (i.e. world comm rank) */
    MPIR_Unsetbit_in_LBI_mask(LBI_mask, my_neighbors_LBI_bit);


    /* Creating INTRA Communicator consisting of my neighbors */
    mpi_errno = MPIR_Comm_create( &(neighborscomm_ptr[my_neighborscomm_rank]) );
    if (mpi_errno) goto fn_fail;
    MPIR_Populate_comm( comm_ptr, neighborscomm_ptr[my_neighborscomm_rank], neighbors_array, size, my_neighbors_CID, my_neighborscomm_rank, num_neighbors, CREATE_RTWMAP /* Create rtw_map */ );
    neighborscomms[my_neighborscomm_rank] = neighborscomm_ptr[my_neighborscomm_rank]->handle;
    
    for (i=0; i<num_neighbors; i++) {
        MPI_Status status;
        int send_to_neighbor[MSGS_TO_NEIGHBOR];
        int recv_from_neighbor[MSGS_TO_NEIGHBOR];
        
        if ( NEED_SORTING ){
            neighbor_rank = sorted_neighbors[i].outercomm_rank;
            neighbor_newLocalRank = sorted_neighbors[i].new_localrank;
        } else {
            neighbor_rank = neighbors_array[i];
            neighbor_newLocalRank = i;            
        }
        
        if (neighbor_rank == fgrank) {
            continue; /* Skipping self */
        }

        /* Sending my neighbor the context-id of my neighbors' communicator
           and my neighbor's rank in that communciator and totprocs */
        send_to_neighbor[0] = my_neighbors_CID; /* context-id of my neighbors' communicator */
        send_to_neighbor[1] = neighbor_newLocalRank; /* The neighbor's local fgrank in my neighbors' communicator */
        send_to_neighbor[2] = num_neighbors; /* totprocs in my neighbors' communicator */
        
        
        if (neighbor_rank > fgrank ){
            MPIC_Send(send_to_neighbor, MSGS_TO_NEIGHBOR, MPI_INT, neighbor_rank, MPIR_COMMS_CREATE_TAG, comm);
            MPIC_Recv(recv_from_neighbor, MSGS_TO_NEIGHBOR, MPI_INT, neighbor_rank, MPIR_COMMS_CREATE_TAG, comm, &status);    
        } else {
            MPIC_Recv(recv_from_neighbor, MSGS_TO_NEIGHBOR, MPI_INT, neighbor_rank, MPIR_COMMS_CREATE_TAG, comm, &status);    
            MPIC_Send(send_to_neighbor, MSGS_TO_NEIGHBOR, MPI_INT, neighbor_rank, MPIR_COMMS_CREATE_TAG, comm);
        }

        if (Is_within_same_HWP(neighbor_rank, comm_ptr, NULL) ) { /* neighbor is collocated */
            /* Creating INTRA Communicator consisting of my neighbor's neighbors */
            mpi_errno = MPIR_Comm_create( &(neighborscomm_ptr[neighbor_newLocalRank]) );
            if (mpi_errno) goto fn_fail;
            MPIR_Populate_comm( comm_ptr, neighborscomm_ptr[neighbor_newLocalRank], NULL, size, recv_from_neighbor[0], recv_from_neighbor[1], recv_from_neighbor[2], SHARE_RTWMAP /* Share rtw_map */ ); 
            neighborscomms[neighbor_newLocalRank] = neighborscomm_ptr[neighbor_newLocalRank]->handle;           
        }
        else { /* neighbor is not collocated so need to get the neighbor's neighbors array */
            int num_neighbor_neighbors = recv_from_neighbor[2];
            int *recv_neighbor_neighbors = (int *)MPIU_Malloc(sizeof(int)*num_neighbor_neighbors);

            if (neighbor_rank > fgrank ){
                MPIC_Send(neighbors_array, num_neighbors, MPI_INT, neighbor_rank, MPIR_COMMS_CREATE_TAG, comm);
                MPIC_Recv(recv_neighbor_neighbors, num_neighbor_neighbors, MPI_INT, neighbor_rank, MPIR_COMMS_CREATE_TAG, comm, &status);    
            } else {
                MPIC_Recv(recv_neighbor_neighbors, num_neighbor_neighbors, MPI_INT, neighbor_rank, MPIR_COMMS_CREATE_TAG, comm, &status);    
                MPIC_Send(neighbors_array, num_neighbors, MPI_INT, neighbor_rank, MPIR_COMMS_CREATE_TAG, comm);
            }

            /* Creating INTRA Communicator consisting of my neighbor's neighbors */
            mpi_errno = MPIR_Comm_create( &(neighborscomm_ptr[neighbor_newLocalRank]) );
            if (mpi_errno) { MPIU_Free(recv_neighbor_neighbors); goto fn_fail; }
            MPIR_Populate_comm( comm_ptr, neighborscomm_ptr[neighbor_newLocalRank], recv_neighbor_neighbors, size, recv_from_neighbor[0], recv_from_neighbor[1], recv_from_neighbor[2], CREATE_OR_SHARE_RTWMAP /* Create rtw_map if not present else share */ );
            neighborscomms[neighbor_newLocalRank] = neighborscomm_ptr[neighbor_newLocalRank]->handle;
            MPIU_Free(recv_neighbor_neighbors);
            
        }
    }
    
    /* ... end of body of routine ... */
    
 fn_exit:
    MPIU_CHKLMEM_FREEALL();
    MPID_MPI_FUNC_EXIT(MPID_STATE_MPIX_COMMS_CREATE);
    return mpi_errno;
    
 fn_fail:
    /* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
    {
	mpi_errno = MPIR_Err_create_code(
                                         mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**mpix_comms_create",
                                         "**mpix_comms_create %C %p", comm, neighborscomms);
    }
#   endif
    mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
    goto fn_exit;
    /* --END ERROR HANDLING-- */
}


#endif /* HK: endif of #if defined(USE_LID_LBI_CONTEXT) */
