/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  $Id: commutil.c,v 1.80 2007/05/04 17:16:15 gropp Exp $
 *
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

#include "mpiimpl.h"
#include "mpicomm.h"


/* HK: Corresponding macro in mpiimpl.h */
#if !defined(FINEGRAIN_MPI)
MPID_Comm MPID_Comm_builtin[MPID_COMM_N_BUILTIN] = { {0} }; /* HK: Separated per FGP. Pointed to by MPIR_Process.comm_world */
#endif
/* This is the utility file for comm that contains the basic comm items
   and storage management */
#ifndef MPID_COMM_PREALLOC 
#define MPID_COMM_PREALLOC 8
#endif
/* Preallocated comm objects */
MPID_Comm MPID_Comm_direct[MPID_COMM_PREALLOC] = { {0} };
MPIU_Object_alloc_t MPID_Comm_mem = { 0, 0, 0, 0, MPID_COMM,
				      sizeof(MPID_Comm), MPID_Comm_direct,
                                      MPID_COMM_PREALLOC};



/* FIXME :
   Reusing context ids can lead to a race condition if (as is desirable)
   MPI_Comm_free does not include a barrier.  Consider the following:
   Process A frees the communicator.
   Process A creates a new communicator, reusing the just released id
   Process B sends a message to A on the old communicator.
   Process A receives the message, and believes that it belongs to the
   new communicator.
   Process B then cancels the message, and frees the communicator.

   The likelyhood of this happening can be reduced by introducing a gap
   between when a context id is released and when it is reused.  An alternative
   is to use an explicit message (in the implementation of MPI_Comm_free)
   to indicate that a communicator is being freed; this will often require
   less communication than a barrier in MPI_Comm_free, and will ensure that 
   no messages are later sent to the same communicator (we may also want to
   have a similar check when building fault-tolerant versions of MPI).
 */

/* Create a new communicator with a context.  
   Do *not* initialize the other fields except for the reference count.
   See MPIR_Comm_copy for a function to produce a copy of part of a
   communicator 
*/


/*  
    Create a communicator structure and perform basic initialization 
    (mostly clearing fields and updating the reference count).  
 */
#undef FUNCNAME
#define FUNCNAME MPIR_Comm_create
#undef FCNAME
#define FCNAME "MPIR_Comm_create"
int MPIR_Comm_create( MPID_Comm **newcomm_ptr )
{   
    int mpi_errno = MPI_SUCCESS;
    MPID_Comm *newptr;
    MPID_MPI_STATE_DECL(MPID_STATE_MPIR_COMM_CREATE);

    MPID_MPI_FUNC_ENTER(MPID_STATE_MPIR_COMM_CREATE);

    newptr = (MPID_Comm *)MPIU_Handle_obj_alloc( &MPID_Comm_mem );
    /* --BEGIN ERROR HANDLING-- */
    if (!newptr) {
	mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, 
		   FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
	goto fn_fail;
    }
    /* --END ERROR HANDLING-- */
    *newcomm_ptr = newptr;
    MPIU_Object_set_ref( newptr, 1 );

    /* Clear many items (empty means to use the default; some of these
       may be overridden within the communicator initialization) */
    newptr->errhandler   = 0;
    newptr->attributes	 = 0;
    newptr->remote_group = 0;
    newptr->local_group	 = 0;
    newptr->coll_fns	 = 0; /* HK: Note that coll_fns are set to null. */
    newptr->topo_fns	 = 0;
    newptr->name[0]	 = 0;

#if defined(FINEGRAIN_MPI)
    newptr->p_rank = -1;
    newptr->totprocs = -1;
    newptr->co_shared_vars = NULL;
    newptr->coFGP_comm = NULL;
    newptr->pid_comm = NULL;
    newptr->isRepresentative = 0;
    newptr->numofcoFGPs = 0;
    newptr->numPidsInComm = 0;
    newptr->barrier_vars = NULL;    
#endif
    /* Fields not set include context_id, remote and local size, and 
       kind, since different communicator construction routines need 
       different values */

    /* Insert this new communicator into the list of known communicators.
       Make this conditional on debugger support to match the test in 
       MPIR_Comm_release . */
    MPIR_COMML_REMEMBER( newptr );

 fn_fail:
    MPID_MPI_FUNC_EXIT(MPID_STATE_MPIR_COMM_CREATE);

    return mpi_errno;
}

/* Create a local intra communicator from the local group of the 
   specified intercomm. */
/* FIXME : 
   For the context id, use the intercomm's context id + 2.  (?)
 */
#undef FUNCNAME
#define FUNCNAME MPIR_Setup_intercomm_localcomm
#undef FCNAME
#define FCNAME "MPIR_Setup_intercomm_localcomm"
int MPIR_Setup_intercomm_localcomm( MPID_Comm *intercomm_ptr )
{
    MPID_Comm *localcomm_ptr;
    int mpi_errno = MPI_SUCCESS;
    MPID_MPI_STATE_DECL(MPID_STATE_MPIR_SETUP_INTERCOMM_LOCALCOMM);

    MPID_MPI_FUNC_ENTER(MPID_STATE_MPIR_SETUP_INTERCOMM_LOCALCOMM);

    localcomm_ptr = (MPID_Comm *)MPIU_Handle_obj_alloc( &MPID_Comm_mem );
    /* --BEGIN ERROR HANDLING-- */
    if (!localcomm_ptr) {
	mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, 
					  FCNAME, __LINE__,
					  MPI_ERR_OTHER, "**nomem", 0 );
	goto fn_fail;
    }
    /* --END ERROR HANDLING-- */
    MPIU_Object_set_ref( localcomm_ptr, 1 );
    /* Note that we must not free this context id since we are sharing it
       with the intercomm's context */
    /* FIXME: This was + 2 (in agreement with the docs) but that
       caused some errors with an apparent use of the same context id
       by operations in different communicators.  Switching this to +1
       seems to have fixed that problem, but this isn't the right answer. */
/*    printf( "intercomm context ids; %d %d\n",
      intercomm_ptr->context_id, intercomm_ptr->recvcontext_id ); */
    /* We use the recvcontext id for both contextids for the localcomm 
     because the localcomm is an intra (not inter) communicator */
    localcomm_ptr->context_id	  = intercomm_ptr->recvcontext_id + 1;
    localcomm_ptr->recvcontext_id = intercomm_ptr->recvcontext_id + 1;

    /* Duplicate the VCRT references */
    MPID_VCRT_Add_ref( intercomm_ptr->local_vcrt );
    localcomm_ptr->vcrt = intercomm_ptr->local_vcrt;
    localcomm_ptr->vcr  = intercomm_ptr->local_vcr;

    /* Save the kind of the communicator */
    localcomm_ptr->comm_kind   = MPID_INTRACOMM;
    
    /* Set the sizes and ranks */
    localcomm_ptr->remote_size = intercomm_ptr->local_size;
    localcomm_ptr->local_size  = intercomm_ptr->local_size;
    localcomm_ptr->rank        = intercomm_ptr->rank;

    /* More advanced version: if the group is available, dup it by 
       increasing the reference count */
    localcomm_ptr->local_group  = 0;
    localcomm_ptr->remote_group = 0;

    /* This is an internal communicator, so ignore */
    localcomm_ptr->errhandler = 0;
    
    /* FIXME  : No local functions for the collectives */
    localcomm_ptr->coll_fns = 0;

    /* FIXME  : No local functions for the topology routines */
    localcomm_ptr->topo_fns = 0;

    /* We do *not* inherit any name */
    localcomm_ptr->name[0] = 0;

    localcomm_ptr->attributes = 0;

    intercomm_ptr->local_comm = localcomm_ptr;

 fn_fail:
    MPID_MPI_FUNC_EXIT(MPID_STATE_MPIR_SETUP_INTERCOMM_LOCALCOMM);

    return mpi_errno;;
}
/*
 * Here are the routines to find a new context id.  The algorithm is discussed 
 * in detail in the mpich2 coding document.  There are versions for
 * single threaded and multithreaded MPI.
 * 
 * These assume that int is 32 bits; they should use uint_32 instead, 
 * and an MPI_UINT32 type (should be able to use MPI_INTEGER4)
 */

/* Both the threaded and non-threaded routines use the same mask of available
   context id values. */

#if !defined(FINEGRAIN_MPI)
#define MAX_CONTEXT_MASK 32
static unsigned int context_mask[MAX_CONTEXT_MASK];
static int initialize_context_mask = 1;
#endif

#ifdef USE_DBG_LOGGING
/* Create a string that contains the context mask.  This is
   used only with the logging interface, and must be used by one thread at 
   a time (should this be enforced by the logging interface?).
   Converts the mask to hex and returns a pointer to that string */
static char *MPIR_ContextMaskToStr( void )
{
    /* HK: TODO A static array of size greater than 32768 cannot be declared on this setup.
       Declaring a dynamic array poses the problem as to where the memory is to be released.
       Also, note that the following static array bufstr is declared as static, so will be
       shared among proclets, but since it is being used in printing a debug message
       only, that won't be a problem.
     */
    static char bufstr[MAX_CONTEXT_MASK*8+1];
    int i;
    int maxset=0;
    for (maxset=MAX_CONTEXT_MASK-1; maxset>=0; maxset--) {
	if (context_mask[maxset] != 0) break;
    }

    for (i=0; i<maxset; i++) {
	MPIU_Snprintf( &bufstr[i*8], 9, "%.8x", context_mask[i] );
    }
    return bufstr;
}
#endif


#if defined(FINEGRAIN_MPI)


#define LBI_UpperMask 0xFFFF0000
#define LBI_LowerMask 0x0000FFFF
#define LBI_SHIFT 16
#define CID_LID_Mask 0xFFFFFC00
#define CID_LBI_Mask 0x000003F0
#define CID_SHIFT 4
#define CID_LID_SHIFT 10 /* k=6 + CID_SHIFT */

inline void MPIR_Get_LBI_count(unsigned int *lbi_mask, int *lbi_count) /* IN, OUT */
{
    int i;
    (*lbi_count) = 0;
    for(i=0; i<MAX_LBI_MASK; i++){
        (*lbi_count) += BitsSetTable256[lbi_mask[i] & 0xff] + 
            BitsSetTable256[(lbi_mask[i] >> 8) & 0xff] + 
            BitsSetTable256[(lbi_mask[i] >> 16) & 0xff] + 
            BitsSetTable256[lbi_mask[i] >> 24]; 
    }   
}

inline int MPIR_Find_LBI_bit( unsigned int *lbi_mask ) {
    int i, j, bitloc = -1;
    for (i=0; i<MAX_LBI_MASK; i++) {
	if (lbi_mask[i]) {
	    /* There is a bit set in this word. */
	    register unsigned int val, nval;
	    /* The following code finds the lowest set bit by recursively
	       checking the bottom half of a subword for a bit, and decrementing
	       the bit location by the number of bit of the higher sub word if 
	       the low subword contains a set bit.  The assumption is that
	       full-word bitwise operations and compares against zero are 
	       fast */
	    val = lbi_mask[i];
	    j   = 31; /* HK: Note: This assumes a 32-bit integer */
	    nval = val & 0x0000FFFF;
	    if (nval) {
		j -= 16;
		val = nval;
	    }
	    nval = val & 0x00FF00FF;
	    if (nval) {
		j -= 8;
		val = nval;
	    }
	    nval = val & 0x0F0F0F0F;
	    if (nval) {
		j -= 4;
		val = nval;
	    }
	    nval = val & 0x33333333;
	    if (nval) {
		j -= 2;
		val = nval;
	    }
	    if (val & 0x55555555) { 
		j -= 1;
	    }

            bitloc = (MPIR_CONTEXT_INT_BITS * i + j);
            return bitloc;        
	}
    }
    return (-1);  /* No bit was set */
}

inline int MPIR_Get_count_LBI_sndval(unsigned int *lbi_mask){
    int lbi_count = -1, lbi_bit = -1;    
    int lbi_count_sndval = 0;
    MPIR_Get_LBI_count(lbi_mask, &lbi_count);
    lbi_bit = MPIR_Find_LBI_bit(lbi_mask); /* This is -1 if no bit is set (i.e. lbi_count is zero */
    if (lbi_bit < 0){
        assert(lbi_count == 0);
    }
    lbi_count_sndval = lbi_count << LBI_SHIFT;
    lbi_count_sndval |= (LBI_LowerMask & lbi_bit);
        
    return (lbi_count_sndval);
}


inline void MPIR_Separate_count_LBI(int lbi_count_sndval, short int* lbi_bit, short int* lbi_count){ /* IN, OUT, OUT */
    
    *lbi_bit = lbi_count_sndval & LBI_LowerMask;
    *lbi_count   = (lbi_count_sndval & LBI_UpperMask) >> LBI_SHIFT ;        
}

inline int MPIR_Get_CID(int LID, short int maxLBI){ /* IN, IN */
    int CID;
    CID = LID << CID_LID_SHIFT;
    CID = CID | (maxLBI << CID_SHIFT);
    return (CID);
}

inline void MPIR_Separate_CID_into_LID_LBI(int cid, int* LID_ptr, short int* lbi_ptr){ /* IN, OUT, OUT */
    *LID_ptr = (cid & CID_LID_Mask) >> CID_LID_SHIFT;
    *lbi_ptr = (cid & CID_LBI_Mask) >> CID_SHIFT;
}


inline void MPIR_Unsetbit_in_LBI_mask(unsigned int *lbi_mask, short int chosen_lbi_for_cid){ /* IN, IN */
    int val = chosen_lbi_for_cid;
    int block = val / MPIR_CONTEXT_INT_BITS;
    int index = val % MPIR_CONTEXT_INT_BITS;
    assert (block < MAX_LBI_MASK );
    assert ( lbi_mask[block] & (1<<index) ); /* Asserting that initially this bit is set */
    lbi_mask[block] &= ~(1<<index); /* Unsetting the bit */
}


inline void MPIR_Setbit_in_LBI_mask(unsigned int *lbi_mask, short int chosen_lbi_for_cid){ /* IN, IN */
    int val = chosen_lbi_for_cid;
    int block = val / MPIR_CONTEXT_INT_BITS;
    int index = val % MPIR_CONTEXT_INT_BITS;
    assert (block < MAX_LBI_MASK );
    assert ( !(lbi_mask[block] & (1<<index)) ); /* Asserting that initially this bit is unset i.e. 0 */
    lbi_mask[block] |= (1<<index); /* Setting the bit */
}


void MPIR_Init_LBI (void) {
    assert (LBI_mask == NULL);
    LBI_mask = (unsigned int *)MPIU_Malloc(MAX_LBI_MASK * sizeof(unsigned int));
    assert (LBI_mask != NULL);
    int i;
    for (i=0; i<MAX_LBI_MASK; i++) {
	LBI_mask[i] = 0xFFFFFFFF;
    }
    /* the first three values are already used (comm_world, comm_self,
       and the internal-only copy of comm_world) */
    if(my_fgrank == 0){ /* HK: root*/
        LBI_mask[0] = 0xFFFFFFF8;
    }
    initialize_LBI_mask = 0;
}

/*
  HK: The bitmaps will be exchanged between the FGPs of ranks 0 to size-1 in comm_ptr.
  The local root is the one with rank=0 in comm_ptr.
*/
int MPIR_Get_LID_LBI_contextid( MPID_Comm *comm_ptr, int size ) {
    
#if !defined(USE_LID_LBI_CONTEXT)
    MPIU_Internal_error_printf("Error: This algorithm only works if USE_LID_LBI_CONTEXT is defined. In function %s at line %d. Aborting...\n", __FUNCTION__, __LINE__);
    MPID_Abort(NULL, MPI_SUCCESS, -1, NULL); 
    MPIU_Exit(-1); /* HK: If for some reason MPID_Abort returns, exit here. */
#endif

    int root = 0, mpi_errno = MPI_SUCCESS;
    int packed_count_LBI, i, k, CID = -1;
    MPI_Comm local_comm = comm_ptr->handle;
    int fgrank = comm_ptr->rank;
    int *recvFromAll = NULL;
    short int maxLBI_count = -1;
    short int maxLBI = -1, LBI, LBI_count;    
    int LID = -1, world_id;        
    
    if(fgrank >= size){
        return (-1);  /* The FGPs of ranks size or greater are not part of the
                        new communicator and hence do not participate in CID
                        generation. Returning CID = -1 for them.
                     */
    }
    
    if(initialize_LBI_mask){
        MPIR_Init_LBI();
    }
    assert (LBI_mask != NULL);
    

    packed_count_LBI = MPIR_Get_count_LBI_sndval(LBI_mask);
    
    if ( fgrank == root ){
        recvFromAll = (int*) MPIU_Malloc(size*sizeof(int));
        assert(recvFromAll);
    }

    
/* Root receives the bitmap info from all FGPs of ranks 0 to size-1 in comm_ptr */        
#if 0 /* Uses explicit Irecvs */
    if ( fgrank == root ) {
        MPI_Request *requests = MPIU_Malloc(sizeof(MPI_Request)*size);
        MPI_Status *requeststats = MPIU_Malloc(sizeof(MPI_Status)*size);
        
        /* --BEGIN ERROR HANDLING-- */
        if ((recvFromAll == NULL) || (requests == NULL) || (requeststats == NULL)) {
            MPIU_Internal_error_printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
            MPID_Abort(NULL, MPI_SUCCESS, -1, NULL); 
            MPIU_Exit(-1); /* HK: If for some reason MPID_Abort returns, exit here. */
        }
        /* --END ERROR HANDLING-- */
        

        for(i=1, k=0; i<(size); i++){ /* Note: Starting from i=1, root is not recveing from itself, copying below */ 
            mpi_errno = MPIC_Irecv(recvFromAll+i, 1, MPI_INT, i, MPIR_COMM_COPY_TAG, local_comm, &requests[k]);
            if (mpi_errno != MPI_SUCCESS) goto fn_fail;
            k++;
        }
        mpi_errno = MPI_Waitall(k, requests,  requeststats); 
        if (mpi_errno != MPI_SUCCESS) goto fn_fail;
    }
    else if(fgrank <size){
        /* Send bitmap to root */
        MPIC_Send(&packed_count_LBI, 1, MPI_INT, root, MPIR_COMM_COPY_TAG, local_comm);
    }
#else /* Uses MPI_Gather */
    mpi_errno = MPI_Gather(&packed_count_LBI, 1, MPI_INT, recvFromAll, 1, MPI_INT, root, local_comm );
    if (mpi_errno != MPI_SUCCESS) goto fn_fail;
#endif

    
    if ( fgrank == root ){
        recvFromAll[root] = packed_count_LBI; /* Root copies its own LBI information */

        for( i=0; i<size; i++ ){
            /* Picking the one with highest count of 1's (i.e., least times leader) */
            MPIR_Separate_count_LBI(recvFromAll[i], &LBI, &LBI_count);
            if((maxLBI_count == -1) || (maxLBI_count < LBI_count)){ 
                maxLBI_count = LBI_count;
                maxLBI = LBI;
                RTWPmapFind(comm_ptr->co_shared_vars->rtw_map, i, &world_id, NULL);
                LID = world_id;
            }
        }

        MPIU_Free(recvFromAll);
        /* Root constructs CID */
        assert(LID >=0);
        if(maxLBI_count == 0){
            assert(maxLBI < 0);
            printf("Failed! Too many communicators! LBI_count is zero for all group members.\n");
            MPIU_ERR_CHKANDJUMP(maxLBI_count == 0, mpi_errno, MPI_ERR_OTHER, 
                                "**toomanycomm" );
            exit(-1); /* Just in case the above falls through */
        }else{
            assert( (maxLBI >= 0) && (maxLBI < (MPIR_CONTEXT_INT_BITS*MAX_LBI_MASK)) );
            assert(maxLBI_count > 0);
        }
        CID = MPIR_Get_CID(LID, maxLBI);
    }

    
#if 0
    if ( fgrank == root ){
        /* uses explicit Isends - replaced with MPI_Bcast below */
        /* Root sends new CID to all FGPs of ranks 0 to size-1 in comm_ptr */
        
        for(i=1, k=0; i<(size); i++){ /* Note: Starting from i=1, root is not sending to itself, copying below */ 
            mpi_errno = MPIC_Isend(&CID, 1, MPI_INT, i, MPIR_COMM_COPY_TAG, local_comm, &requests[k]);
            if (mpi_errno != MPI_SUCCESS) goto fn_fail;
            k++;
        }

        mpi_errno = MPI_Waitall(k, requests,  requeststats); 
        if (mpi_errno != MPI_SUCCESS) goto fn_fail;

        
        if(requests) MPIU_Free(requests);
        if(requeststats) MPIU_Free(requeststats);
    }
    else if (fgrank <size ){
        /* Receive CID from root */
        MPI_Status recvstatus;
        MPIC_Recv( &CID, 1, MPI_INT, root, MPIR_COMM_COPY_TAG, local_comm, &recvstatus );
    }
    
#else /* Uses MPI_Bcast */
    
    mpi_errno = MPI_Bcast(&CID, 1, MPI_INT, root, local_comm );
    if (mpi_errno != MPI_SUCCESS) goto fn_fail;
    
#endif
        

    /* Update of LBI mask by the chosen CID leader */
    int chosen_LID_for_CID = -1;
    short int chosen_LBI_for_CID = -1;
    MPIR_Separate_CID_into_LID_LBI(CID, &chosen_LID_for_CID, &chosen_LBI_for_CID);
    assert( (chosen_LID_for_CID >=0) && (chosen_LBI_for_CID >=0) &&
            (chosen_LBI_for_CID < (MPIR_CONTEXT_INT_BITS*MAX_LBI_MASK)) );             
    /* Remember chosen_LID_for_CID is a worldrank, so comparing against my_fgrank */
    if (my_fgrank == chosen_LID_for_CID) {
        /* unset the bit location in chosen_LBI_for_CID */
        MPIR_Unsetbit_in_LBI_mask(LBI_mask, chosen_LBI_for_CID);
    }            

 fn_exit:
    assert(CID != -1);
    return (CID);

 fn_fail:
    MPIU_Internal_error_printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Arrived at fn_fail....\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
    MPID_Abort(NULL, MPI_SUCCESS, -1, NULL); 
    MPIU_Exit(-1); /* HK: If for some reason MPID_Abort returns, exit here. */        
}

void MPIR_Intra_newcomm_init(MPID_Comm *newcomm_ptr, int pid_rank, int new_totprocs)
{
    /* NOTE: This function assumes that newcomm_ptr->remote_size, newcomm_ptr->rank (i.e. finegrain rank),
       newcomm_ptr->context_id and newcomm_ptr->recvcontext_id are being set elsewhere.
    */ 
    newcomm_ptr->p_rank         = pid_rank; /* local rank in new communicator  _p_rank_ */
    newcomm_ptr->totprocs	= new_totprocs; 
    
    newcomm_ptr->coFGP_comm     = NULL;
    newcomm_ptr->pid_comm       = NULL;
    newcomm_ptr->isRepresentative = 0;
    newcomm_ptr->numofcoFGPs    = 0;
    newcomm_ptr->numPidsInComm  = 0;        
    newcomm_ptr->barrier_vars   = NULL;

    newcomm_ptr->vcrt           = vcrt_world;
    newcomm_ptr->vcr            = vcr_world;
}

                 
int MPIR_Copy_and_share_rtwmap(MPID_Comm *newcomm_ptr, MPID_Comm *comm_ptr)
{
    /* HK:
       NOTE: This function should be called after MPIR_Intra_newcomm_init()
       from inside MPIR_Comm_copy() and and will always make a copy of
       the outer comm rtw map
       
       Looking up pointer to shared rtw map in contextLeader_hshtbl.
       If it does not exist, then the first proclet to enter this
       function creates it.
    */
    int i, mpi_errno = MPI_SUCCESS;
    int fgrank = newcomm_ptr->rank;
    int new_totprocs = newcomm_ptr->totprocs;
    cidLookupHashItemptr stored = NULL;
    CidLookupHashFind(cidLookup_hshtbl, newcomm_ptr->context_id, &stored);
    if (stored) {
        MPIR_Comm_add_coshared_all_ref(stored->coproclet_shared_vars);
        newcomm_ptr->co_shared_vars = ((Coproclet_shared_vars_t *)(stored->coproclet_shared_vars));
        MPIR_Copy_ch3i_comm_ch(newcomm_ptr, ((Coproclet_shared_vars_t *)(stored->coproclet_shared_vars))->ch3i_ptr);
        FGP_DBG_PRINTF("Already present in cidLookup_hshtbl, context-id=%d, newcomm_ptr->co_shared_vars->rtw_map=%p\n", newcomm_ptr->context_id, newcomm_ptr->co_shared_vars->rtw_map);
    }
    else {
        /* HK: If  new_totprocs == comm_ptr->totprocs then share the RTWmap
           with the original (i.e. the one being duplicated).           
        */
        assert(stored == NULL);
        newcomm_ptr->co_shared_vars = (Coproclet_shared_vars_t *)MPIU_Malloc(sizeof(Coproclet_shared_vars_t));
        /* --BEGIN ERROR HANDLING-- */
        if ( (newcomm_ptr->co_shared_vars == NULL) ) {                    
            printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
            mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
            goto fn_fail;
        }
        /* --END ERROR HANDLING-- */
        
        if (new_totprocs == comm_ptr->totprocs)
        {
            newcomm_ptr->co_shared_vars->rtw_map = comm_ptr->co_shared_vars->rtw_map;
        } else {
            newcomm_ptr->co_shared_vars->rtw_map = (RTWmap *)RTWmapCreate(new_totprocs);
            assert( newcomm_ptr->co_shared_vars->rtw_map != NULL );
            int *rtw_blockinsert = (int*) MPIU_Malloc(sizeof(int) * new_totprocs);
            assert(rtw_blockinsert != 0);
            for (i=0; i<new_totprocs; i++) {
                /* Initializing worldrank to -1 */
                int worldrank = -1, key;
                key = i; /* HK: This is the local fgrank */
                RTWPmapFind(comm_ptr->co_shared_vars->rtw_map, key, &worldrank, NULL);
                /* HK: Store item in newcomm_ptr rtw map which has i as
                   the local fgrank relative to the new communicator.
                   The corresponding worldrank in outer communicator is
                   determined and saved against fgrank=i in the newcomm rtw map. 
                */
                rtw_blockinsert[key] = worldrank;                    
            }
                
            RTWmapBlockInsert(newcomm_ptr->co_shared_vars->rtw_map, new_totprocs, rtw_blockinsert);
            MPIU_Free(rtw_blockinsert);
        }
        
        newcomm_ptr->co_shared_vars->co_barrier_vars =  (struct coproclet_barrier_vars *)MPIU_Malloc(sizeof(struct coproclet_barrier_vars));
 
        /* --BEGIN ERROR HANDLING-- */
        if ( (newcomm_ptr->co_shared_vars->co_barrier_vars == NULL) ) {                    
            printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
            mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
            goto fn_fail;
        }
        /* --END ERROR HANDLING-- */
                
        newcomm_ptr->co_shared_vars->co_barrier_vars->coproclet_signal = 0;
        newcomm_ptr->co_shared_vars->co_barrier_vars->coproclet_counter = 0;
        newcomm_ptr->co_shared_vars->co_barrier_vars->leader_signal = 0;

        /* Notify the device of the new communicator */
        MPID_Dev_comm_create_hook(newcomm_ptr);

        if (new_totprocs == comm_ptr->totprocs)
        {
            stored = NULL;
            CidLookupHashFind(cidLookup_hshtbl, comm_ptr->context_id, &stored); 
            assert(stored != NULL);
            newcomm_ptr->co_shared_vars->ref_acrossComm_countptr = ((Coproclet_shared_vars_t *)(stored->coproclet_shared_vars))->ref_acrossComm_countptr;
            MPIR_Comm_add_coshared_acrossComm_ref(newcomm_ptr->co_shared_vars); 
            MPIR_Comm_init_coshared_withinComm_ref(newcomm_ptr->co_shared_vars);
        } else {
            MPIR_Comm_init_coshared_all_ref(newcomm_ptr->co_shared_vars); 
        }
        newcomm_ptr->co_shared_vars->ch3i_ptr = &(newcomm_ptr->ch); /* HK: TODO IMPORTANT when ch3i_ptr is also shared */
        stored = NULL;
        CidLookupHashInsert(cidLookup_hshtbl, newcomm_ptr->context_id, newcomm_ptr->co_shared_vars, &stored);
        if(!stored){
            printf("Store failure in contextLeader_hshtbl\n");
            MPIU_ERR_CHKANDJUMP(stored == 0, mpi_errno, MPI_ERR_OTHER, 
                                "**hshstorefail" );
            exit(-1);
        }
        FGP_DBG_PRINTF("Inserting in cidLookup_hshtbl, context-id=%d, newcomm_ptr->co_shared_vars->rtw_map=%p\n", newcomm_ptr->context_id, newcomm_ptr->co_shared_vars->rtw_map);
    }
    
 fn_fail:
 fn_exit:
    return(mpi_errno);
    
}

inline void MPIR_Coshared_group_release(MPID_Group * group_ptr){    
    if ( group_ptr->ref_acrossCommGroup_countptr ) { /* HK: !NULL, i.e. a communicator has been created using this group */
        assert( group_ptr->rtw_grp_map != NULL );
        MPIR_Comm_release_coshared_group_ref(group_ptr);
        FGP_DBG_PRINTF("my_fgrank=%d...................... *(group_ptr->ref_acrossCommGroup_countptr)=%d\n", my_fgrank, *group_ptr->ref_acrossCommGroup_countptr);
        if(!(*(group_ptr->ref_acrossCommGroup_countptr))){ /* HK: No communicator is sharing
                                                              rtw_grp_map,i.e. MPIR_Comm_add_coshared_group_ref()
                                                              has never been called. */
            MPIU_Free(group_ptr->ref_acrossCommGroup_countptr);
            RTWmapKill(&(group_ptr->rtw_grp_map));
            assert(group_ptr->rtw_grp_map == NULL);
        }
    } else if(group_ptr->rtw_grp_map != NULL){ /* HK: This would be the case when e.g. MPI_Group_incl() is used
                                                  to create a group and then MPI_Group_free() is called without
                                                  creating a communicator first. */
        RTWmapKill(&(group_ptr->rtw_grp_map));
        assert(group_ptr->rtw_grp_map == NULL);        
    }
}



#endif /* HK: (FINEGRAIN_MPI) */



static void MPIR_Init_contextid (void) {
#if defined(FINEGRAIN_MPI)
    assert (context_mask == NULL);
    context_mask = (unsigned int *)MPIU_Malloc(MAX_CONTEXT_MASK * sizeof(unsigned int));
#endif
    int i;
    for (i=1; i<MAX_CONTEXT_MASK; i++) {
	context_mask[i] = 0xFFFFFFFF;
    }
    /* the first three values are already used (comm_world, comm_self,
       and the internal-only copy of comm_world) */
    context_mask[0] = 0xFFFFFFF8; 
    initialize_context_mask = 0;
}
/* Return the context id corresponding to the first set bit in the mask.
   Return 0 if no bit found */
static int MPIR_Find_context_bit( unsigned int local_mask[] ) {
    int i, j, context_id = 0;
    for (i=0; i<MAX_CONTEXT_MASK; i++) {
	if (local_mask[i]) {
	    /* There is a bit set in this word. */
	    register unsigned int val, nval;
	    /* The following code finds the highest set bit by recursively
	       checking the top half of a subword for a bit, and incrementing
	       the bit location by the number of bit of the lower sub word if 
	       the high subword contains a set bit.  The assumption is that
	       full-word bitwise operations and compares against zero are 
	       fast */
	    val = local_mask[i];
	    j   = 0;
	    nval = val & 0xFFFF0000;
	    if (nval) {
		j += 16;
		val = nval;
	    }
	    nval = val & 0xFF00FF00;
	    if (nval) {
		j += 8;
		val = nval;
	    }
	    nval = val & 0xF0F0F0F0;
	    if (nval) {
		j += 4;
		val = nval;
	    }
	    nval = val & 0xCCCCCCCC;
	    if (nval) {
		j += 2;
		val = nval;
	    }
	    if (val & 0xAAAAAAAA) { 
		j += 1;
	    }
	    context_mask[i] &= ~(1<<j);
            context_id = 4 * (32 * i + j);            
	    MPIU_DBG_MSG_FMT(COMM,VERBOSE,(MPIU_DBG_FDEST,
                    "allocating contextid = %d, (mask[%d], bit %d\n", 
		    context_id, i, j ) ); 
	    return context_id;
	}
    }
    return 0;
}
#ifndef MPICH_IS_THREADED
/* Unthreaded (only one MPI call active at any time) */

#undef FUNCNAME
#define FUNCNAME MPIR_Get_contextid
#undef FCNAME
#define FCNAME "MPIR_Get_contextid"
/* HK: For single threaded case. Note this function is **NOT** being called,
   but the one below for MPICH_IS_THREADED is set and true, is being called. */
int MPIR_Get_contextid( MPID_Comm *comm_ptr ) 
{
    int          context_id = 0;
    int          mpi_errno = 0;
    unsigned int local_mask[MAX_CONTEXT_MASK];
    MPIU_THREADPRIV_DECL;
    MPID_MPI_STATE_DECL(MPID_STATE_MPIR_GET_CONTEXTID);

    MPID_MPI_FUNC_ENTER(MPID_STATE_MPIR_GET_CONTEXTID);

    if (initialize_context_mask) { 
	MPIR_Init_contextid();
    }
    memcpy( local_mask, context_mask, MAX_CONTEXT_MASK * sizeof(int) );

    MPIU_THREADPRIV_GET;
    MPIR_Nest_incr();

    FGP_DBG_PRINTF("single threaded case. Before NMPI_Allreduce\n");
    /* Comm must be an intracommunicator */
    mpi_errno = NMPI_Allreduce( MPI_IN_PLACE, local_mask, MAX_CONTEXT_MASK, 
				MPI_INT, MPI_BAND, comm_ptr->handle ); 
    MPIR_Nest_decr();
    /* FIXME: We should return the error code upward */
    if (mpi_errno) {
	MPID_MPI_FUNC_EXIT(MPID_STATE_MPIR_GET_CONTEXTID);
	return 0;
    }

    context_id = MPIR_Find_context_bit( local_mask );
    FGP_DBG_PRINTF("single threaded case. After NMPI_Allreduce context_id=%d\n", context_id);
    
    MPIU_DBG_MSG_S(COMM,VERBOSE,"Context mask = %s",MPIR_ContextMaskToStr());

    /* return 0 if no context id found.  The calling routine should 
       check for this and generate the appropriate error code */
    MPID_MPI_FUNC_EXIT(MPID_STATE_MPIR_GET_CONTEXTID);

    return context_id;
}

#else /* MPICH_IS_THREADED is set and true */


#if !defined(FINEGRAIN_MPI)
#define MPIR_MAXID (1 << 30)
/* Additional values needed to maintain thread safety */
static volatile int mask_in_use = 0;
/* lowestContextId is used to break ties when multiple threads
   are contending for the mask */
static volatile int lowestContextId = MPIR_MAXID;
#endif


#undef FUNCNAME
#define FUNCNAME MPIR_Get_contextid
#undef FCNAME
#define FCNAME "MPIR_Get_contextid"
/* HK: Note this is when MPICH_IS_THREADED is set and true. This function IS being called. */
int MPIR_Get_contextid( MPID_Comm *comm_ptr ) 
{
    int          context_id = 0;
    int          mpi_errno = 0;
#if defined(FINEGRAIN_MPI)
    unsigned int *local_mask = (unsigned int *)MPIU_Malloc(MAX_CONTEXT_MASK * sizeof(unsigned int));
    MPIU_Assert (local_mask != 0);
#else
    unsigned int local_mask[MAX_CONTEXT_MASK];
#endif
    int          own_mask = 0;
    int          testCount = 10;

    MPIU_THREADPRIV_DECL;
    MPID_MPI_STATE_DECL(MPID_STATE_MPIR_GET_CONTEXTID);

    MPID_MPI_FUNC_ENTER(MPID_STATE_MPIR_GET_CONTEXTID);

    MPIU_THREADPRIV_GET;

    /* We increment the nest level now because we need to know that we're
     within another MPI routine before calling the CS_ENTER macro */
    MPIR_Nest_incr();

    /* The SINGLE_CS_ENTER/EXIT macros are commented out because this
       routine shold always be called from within a routine that has 
       already entered the single critical section.  However, in a 
       finer-grained approach, these macros indicate where atomic updates
       to the shared data structures must be protected. */
    /* We lock only around access to the mask.  If another thread is
       using the mask, we take a mask of zero */
    MPIU_DBG_MSG_FMT( COMM, VERBOSE, (MPIU_DBG_FDEST,
         "Entering; shared state is %d:%d", mask_in_use, lowestContextId ) );
    /* We need a special test in this loop for the case where some process
     has exhausted its supply of context ids.  In the single threaded case, 
     this is simple, because the algorithm is deterministic (see above).  In 
     the multithreaded case, it is more complicated, because we may get a
     zero for the context mask because some other thread holds the mask.  
     In addition, we can't check for the case where this process did not
     select MPI_THREAD_MULTIPLE, because one of the other processes
     may have selected MPI_THREAD_MULTIPLE.  To handle this case, after a 
     fixed number of failures, we test to see if some process has exhausted 
     its supply of context ids.  If so, all processes can invoke the 
     out-of-context-id error.  That fixed number of tests is in testCount */
    while (context_id == 0) {
	/* MPIU_THREAD_SINGLE_CS_ENTER("context_id"); */
	if (initialize_context_mask) { /* HK: context_mask[] & initialize_context_mask defined per FGP */
	    MPIR_Init_contextid();
	}
	if (mask_in_use || comm_ptr->context_id > lowestContextId) { /* HK: mask_in_use & lowestContextId defined per FGP */
	    memset( local_mask, 0, MAX_CONTEXT_MASK * sizeof(int) );
	    own_mask        = 0;
	    if (comm_ptr->context_id < lowestContextId) {
		lowestContextId = comm_ptr->context_id;
	    }
	    MPIU_DBG_MSG_D( COMM, VERBOSE, 
	       "In in-use, sed lowestContextId to %d", lowestContextId );
	}
	else {
	    memcpy( local_mask, context_mask, MAX_CONTEXT_MASK * sizeof(int) );
	    mask_in_use     = 1;
	    own_mask        = 1;
	    lowestContextId = comm_ptr->context_id;
	    MPIU_DBG_MSG( COMM, VERBOSE, "Copied local_mask" );
	}
	/* MPIU_THREAD_SINGLE_CS_EXIT("context_id"); */

	/* Now, try to get a context id */
	/* Comm must be an intracommunicator */
	mpi_errno = NMPI_Allreduce( MPI_IN_PLACE, local_mask, MAX_CONTEXT_MASK,
				    MPI_INT, MPI_BAND, comm_ptr->handle );
	/* FIXME: On error, return mpi_errno upward */
	if (mpi_errno) {
	    MPIR_Nest_decr();
	    MPID_MPI_FUNC_EXIT(MPID_STATE_MPIR_GET_CONTEXTID);
	    return 0;
	}

	if (own_mask) {
	    /* There is a chance that we've found a context id */
	    /* MPIU_THREAD_SINGLE_CS_ENTER("context_id"); */
	    /* Find_context_bit updates the context array if it finds a match */
	    context_id = MPIR_Find_context_bit( local_mask );
	    MPIU_DBG_MSG_D( COMM, VERBOSE, 
			    "Context id is now %d", context_id );
	    if (context_id > 0) {
		/* If we were the lowest context id, reset the value to
		   allow the other threads to compete for the mask */
		if (lowestContextId == comm_ptr->context_id) {
		    lowestContextId = MPIR_MAXID;
		    /* Else leave it alone; there is another thread waiting */
		}
	    }
	    else {
		/* else we did not find a context id. Give up the mask in case
		   there is another thread (with a lower context id) waiting for
		   it.
		   We need to ensure that any other threads have the 
		   opportunity to run.  We do this by releasing the single
		   mutex, yielding, and then reaquiring the mutex.
		   We might want to do something more sophisticated, such
		   as using a condition variable (if we know for sure that
		   there is another thread on this process that is waiting).
		*/
		MPID_Thread_mutex_unlock(&MPIR_ThreadInfo.global_mutex);
		MPID_Thread_yield();
		MPID_Thread_mutex_lock(&MPIR_ThreadInfo.global_mutex);
	    }
	    mask_in_use = 0;
	    /* MPIU_THREAD_SINGLE_CS_EXIT("context_id"); */
	}
	else {
	    /* As above, force this thread to yield */
	    MPID_Thread_mutex_unlock(&MPIR_ThreadInfo.global_mutex);
	    MPID_Thread_yield();
	    MPID_Thread_mutex_lock(&MPIR_ThreadInfo.global_mutex);
	}
	/* Here is the test for out-of-context ids */
	if (testCount-- == 0) {
	    int hasNoId, totalHasNoId;
	    /* We don't need to lock on this because we're just looking for
	       zero or nonzero */
	    hasNoId = MPIR_Find_context_bit( context_mask ) == 0;
	    NMPI_Allreduce( &hasNoId, &totalHasNoId, 1, MPI_INT, 
			    MPI_MAX, comm_ptr->handle );
	    if (totalHasNoId == 1) {
		/* Failure */
		context_id = 0;
		/* Release the masks */
		if (own_mask) {
		    mask_in_use = 0;
		}
		break;
	    }
	}
    }

    MPIU_DBG_MSG_S(COMM,VERBOSE,"Context mask = %s",MPIR_ContextMaskToStr());

    MPIR_Nest_decr();
    MPID_MPI_FUNC_EXIT(MPID_STATE_MPIR_GET_CONTEXTID);
#if defined(FINEGRAIN_MPI)
    MPIU_Free(local_mask);
#endif
    
    return context_id;
}
#endif

/* Get a context for a new intercomm.  There are two approaches 
   here (for MPI-1 codes only)
   (a) Each local group gets a context; the groups exchange, and
       the low value is accepted and the high one returned.  This
       works because the context ids are taken from the same pool.
   (b) Form a temporary intracomm over all processes and use that
       with the regular algorithm.
   
   In some ways, (a) is the better approach because it is the one that
   extends to MPI-2 (where the last step, returning the context, is 
   not used and instead separate send and receive context id value 
   are kept).  For this reason, we'll use (a).

   Even better is to separate the local and remote context ids.  Then
   each group of processes can manage their context ids separately.
*/
/* 
 * This uses the thread-safe (if necessary) routine to get a context id
 * and does not need its own thread-safe version.
 */
#undef FUNCNAME
#define FUNCNAME MPIR_Get_intercomm_contextid
#undef FCNAME
#define FCNAME "MPIR_Get_intercomm_contextid"
int MPIR_Get_intercomm_contextid( MPID_Comm *comm_ptr, int *context_id, 
				  int *recvcontext_id )
{
    int mycontext_id, remote_context_id;
    int mpi_errno = MPI_SUCCESS;
    int tag = 31567; /* FIXME  - we need an internal tag or 
		        communication channel.  Can we use a different
		        context instead?.  Or can we use the tag 
		        provided in the intercomm routine? (not on a dup, 
			but in that case it can use the collective context) */
    MPIU_THREADPRIV_DECL;
    MPID_MPI_STATE_DECL(MPID_STATE_MPIR_GET_INTERCOMM_CONTEXTID);

    MPID_MPI_FUNC_ENTER(MPID_STATE_MPIR_GET_INTERCOMM_CONTEXTID);

    if (!comm_ptr->local_comm) {
	/* Manufacture the local communicator */
	MPIR_Setup_intercomm_localcomm( comm_ptr );
    }

    /*printf( "local comm size is %d and intercomm local size is %d\n",
      comm_ptr->local_comm->local_size, comm_ptr->local_size );*/
    mycontext_id = MPIR_Get_contextid( comm_ptr->local_comm );
    if (mycontext_id == 0) {
	goto fn_fail;
    }

    MPIU_THREADPRIV_GET;

    /* MPIC routine uses an internal context id.  The local leads (process 0)
       exchange data */
    remote_context_id = -1;
    if (comm_ptr->rank == 0) {
	MPIC_Sendrecv( &mycontext_id, 1, MPI_INT, 0, tag,
		       &remote_context_id, 1, MPI_INT, 0, tag, 
		       comm_ptr->handle, MPI_STATUS_IGNORE );
    }

    /* Make sure that all of the local processes now have this
       id */
    MPIR_Nest_incr();
    NMPI_Bcast( &remote_context_id, 1, MPI_INT, 
		0, comm_ptr->local_comm->handle );
    MPIR_Nest_decr();
    /* printf( "intercomm context = %d\n", remote_context_id ); */

    *context_id     = remote_context_id;
    *recvcontext_id = mycontext_id;
 fn_fail:
    MPID_MPI_FUNC_EXIT(MPID_STATE_MPIR_GET_INTERCOMM_CONTEXTID);
    return mpi_errno;
}


#undef FUNCNAME
#define FUNCNAME MPIR_Free_contextid
#undef FCNAME
#define FCNAME "MPIR_Free_contextid"

#if defined(USE_LID_LBI_CONTEXT) /* HK: */

void MPIR_Free_contextid( int context_id )
{
    int chosen_LID_for_CID = -1;
    short int chosen_LBI_for_CID = -1;

    MPID_MPI_STATE_DECL(MPID_STATE_MPIR_FREE_CONTEXTID);
    
    MPID_MPI_FUNC_ENTER(MPID_STATE_MPIR_FREE_CONTEXTID);

    /* Update of LBI mask by the chosen CID leader */
    MPIR_Separate_CID_into_LID_LBI(context_id, &chosen_LID_for_CID, &chosen_LBI_for_CID);
    assert( (chosen_LID_for_CID >=0) && (chosen_LBI_for_CID >=0) &&
            (chosen_LBI_for_CID < (MPIR_CONTEXT_INT_BITS*MAX_LBI_MASK)) );             
    /* Remember chosen_LID_for_CID is a worldrank, so comparing against my_fgrank */
    if (my_fgrank == chosen_LID_for_CID) {
        /* set the bit location in chosen_LBI_for_CID */
        MPIR_Setbit_in_LBI_mask(LBI_mask, chosen_LBI_for_CID);
    }
    
    MPIU_DBG_MSG_FMT(COMM,VERBOSE,(MPIU_DBG_FDEST,
			"Freed context %d, chosen_LID_for_CID %d, chosen_LBI_for_CID %d\n", 
                                   context_id, chosen_LID_for_CID, chosen_LBI_for_CID ) );
    MPID_MPI_FUNC_EXIT(MPID_STATE_MPIR_FREE_CONTEXTID);
}

#else /* HK: #else USE_LID_LBI_CONTEXT is not defined, use MPICH2's context_id release algorithm. */

void MPIR_Free_contextid( int context_id )
{
    int idx, bitpos;
    MPID_MPI_STATE_DECL(MPID_STATE_MPIR_FREE_CONTEXTID);
    
    MPID_MPI_FUNC_ENTER(MPID_STATE_MPIR_FREE_CONTEXTID);

    /* Convert the context id to the bit position */
    /* printf( "Freed id = %d\n", context_id ); */
    context_id >>= 2;       /* Remove the shift of a factor of four */
    idx    = context_id / 32;
    bitpos = context_id % 32;

    /* --BEGIN ERROR HANDLING-- */
    if (idx < 0 || idx >= MAX_CONTEXT_MASK) {
	MPID_Abort( 0, MPI_ERR_INTERN, 1, 
		    "In MPIR_Free_contextid, idx is out of range" );
    }
    /* --END ERROR HANDLING-- */

    /* Check that this context id has been allocated */
#if 0
    /* FIXME: This test should be included */
    if ( (context_mask[idx] & (0x1 << bitpos)) != 0) {
	MPID_Abort( 0, MPI_ERR_INTERN, 1, 
		    "In MPIR_Free_contextid, the context id is not in use" );
    }
#endif
    /* MT: Note that this update must be done atomically in the multithreaded
       case.  In the "one, single lock" implementation, that lock is indeed
       held when this operation is called. */
    context_mask[idx] |= (0x1 << bitpos);

    MPIU_DBG_MSG_FMT(COMM,VERBOSE,(MPIU_DBG_FDEST,
			"Freed context %d, mask[%d] bit %d\n", 
			context_id, idx, bitpos ) );
    MPID_MPI_FUNC_EXIT(MPID_STATE_MPIR_FREE_CONTEXTID);
}

#endif /*(USE_LID_LBI_CONTEXT) */

/*
 * Copy a communicator, including creating a new context and copying the
 * virtual connection tables and clearing the various fields.
 * Does *not* copy attributes.  If size is < the size of the (local group
 * in the ) input communicator, copy only the first size elements.  
 * If this process is not a member, return a null pointer in outcomm_ptr.
 * This is only supported in the case where the communicator is in 
 * Intracomm (not an Intercomm).  Note that this is all that is required
 * for cart_create and graph_create.
 *
 * Used by cart_create, graph_create, and dup_create 
 */
#undef FUNCNAME
#define FUNCNAME MPIR_Comm_copy
#undef FCNAME
#define FCNAME "MPIR_Comm_copy"
int MPIR_Comm_copy( MPID_Comm *comm_ptr, int size, MPID_Comm **outcomm_ptr )
{
    int mpi_errno = MPI_SUCCESS;
    int new_context_id, new_recvcontext_id;
    MPID_Comm *newcomm_ptr;
    MPID_MPI_STATE_DECL(MPID_STATE_MPIR_COMM_COPY);

    MPID_MPI_FUNC_ENTER(MPID_STATE_MPIR_COMM_COPY);

    /* Get a new context first.  We need this to be collective over the
       input communicator */
#if defined(FINEGRAIN_MPI)
    new_context_id = MPIR_Get_LID_LBI_contextid( comm_ptr, size );
    FGP_DBG_PRINTF("...........................new_context_id=%d\n", new_context_id);
    new_recvcontext_id = new_context_id;    
#else
    /* If there is a context id cache in oldcomm, use it here.  Otherwise,
       use the appropriate algorithm to get a new context id.  Be careful
       of intercomms here */
    if (comm_ptr->comm_kind == MPID_INTERCOMM) { /* HK: TODO INTERCOMM */
	mpi_errno = 
	    MPIR_Get_intercomm_contextid( 
		 comm_ptr, &new_context_id, &new_recvcontext_id );
    }
    else {
	new_context_id = MPIR_Get_contextid( comm_ptr );
	new_recvcontext_id = new_context_id;
    }
#endif
    
    /* --BEGIN ERROR HANDLING-- */
    if (new_context_id == 0 || mpi_errno != MPI_SUCCESS) {
	mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
               FCNAME, __LINE__, MPI_ERR_OTHER, "**toomanycomm", 0 );
	goto fn_fail;
    }
    /* --END ERROR HANDLING-- */

    /* This is the local size, not the remote size, in the case of
       an intercomm */
    if (comm_ptr->rank >= size) {
	*outcomm_ptr = 0;
	goto fn_exit;
    }

    /* We're left with the processes that will have a non-null communicator.
       Create the object, initialize the data, and return the result */

    mpi_errno = MPIR_Comm_create( &newcomm_ptr );
    if (mpi_errno) {
        mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
                                          FCNAME, __LINE__, MPI_ERR_OTHER, "**mpir_comm_create_failed", 0 );
        goto fn_fail;
    }

    newcomm_ptr->context_id     = new_context_id;
    newcomm_ptr->recvcontext_id = new_recvcontext_id;

#if defined(FINEGRAIN_MPI)
    MPIR_Intra_newcomm_init(newcomm_ptr, comm_ptr->p_rank, size); /* _p_rank_ */
#endif    
    
    /* Save the kind of the communicator */
    newcomm_ptr->comm_kind   = comm_ptr->comm_kind;
    newcomm_ptr->local_comm  = 0;

#if !defined(FINEGRAIN_MPI)
    /* There are two cases here - size is the same as the old communicator,
       or it is smaller.  If the size is the same, we can just add a reference.
       Otherwise, we need to create a new VCRT.  Note that this is the
       test that matches the test on rank above. */
    if (size == comm_ptr->local_size) {
	/* Duplicate the VCRT references */
	MPID_VCRT_Add_ref( comm_ptr->vcrt );
	newcomm_ptr->vcrt = comm_ptr->vcrt;
	newcomm_ptr->vcr  = comm_ptr->vcr;
    }
    else {
	int i;
	/* The "remote" vcr gets the shortened vcrt */
	MPID_VCRT_Create( size, &newcomm_ptr->vcrt );
	MPID_VCRT_Get_ptr( newcomm_ptr->vcrt, 
			   &newcomm_ptr->vcr );
	for (i=0; i<size; i++) {
	    /* For rank i in the new communicator, find the corresponding
	       rank in the input communicator */
	    MPID_VCR_Dup( comm_ptr->vcr[i], &newcomm_ptr->vcr[i] );
	}
    }
#endif

    /* If it is an intercomm, duplicate the local vcrt references */
    if (comm_ptr->comm_kind == MPID_INTERCOMM) { /* HK: TODO INTERCOMM */
	MPID_VCRT_Add_ref( comm_ptr->local_vcrt );
	newcomm_ptr->local_vcrt = comm_ptr->local_vcrt;
	newcomm_ptr->local_vcr  = comm_ptr->local_vcr;
    }

    /* Set the sizes and ranks */
    newcomm_ptr->rank        = comm_ptr->rank;
    if (comm_ptr->comm_kind == MPID_INTERCOMM) { /* HK: TODO INTERCOMM */
	newcomm_ptr->local_size   = comm_ptr->local_size;
	newcomm_ptr->remote_size  = comm_ptr->remote_size;
	newcomm_ptr->is_low_group = comm_ptr->is_low_group;
    }
    else {
#if defined(FINEGRAIN_MPI)
	newcomm_ptr->local_size  = comm_ptr->local_size;
	newcomm_ptr->remote_size = comm_ptr->local_size;
#else
	newcomm_ptr->local_size  = size;
	newcomm_ptr->remote_size = size;
#endif
    }

    /* Inherit the error handler (if any) */
    newcomm_ptr->errhandler = comm_ptr->errhandler;
    if (comm_ptr->errhandler) {
	MPIR_Errhandler_add_ref( comm_ptr->errhandler );
    }

#if defined(FINEGRAIN_MPI)
    /* HK: Sharing of outer comm rtw map is enabled through ref_acrossComm_countptr */
    mpi_errno = MPIR_Copy_and_share_rtwmap(newcomm_ptr, comm_ptr);
    if (mpi_errno) {
        mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE,
                                          FCNAME, __LINE__, MPI_ERR_OTHER, "**mpir_copy_and_share_rtwmap_failed", 0 );        
        goto fn_fail;
    }
#else
    /* Notify the device of the new communicator */
    MPID_Dev_comm_create_hook(newcomm_ptr);
#endif
    
    /* Start with no attributes on this communicator */
    newcomm_ptr->attributes = 0;
    *outcomm_ptr = newcomm_ptr;

 fn_fail:
 fn_exit:

    MPID_MPI_FUNC_EXIT(MPID_STATE_MPIR_COMM_COPY);

    return mpi_errno;
}

/* Release a reference to a communicator.  If there are no pending
   references, delete the communicator and recover all storage and 
   context ids */
#undef FUNCNAME 
#define FUNCNAME MPIR_Comm_release
#undef FCNAME
#define FCNAME "MPIR_Comm_release"
int MPIR_Comm_release(MPID_Comm * comm_ptr, int isDisconnect) 
{
    int mpi_errno = MPI_SUCCESS;
    int inuse;
    MPID_MPI_STATE_DECL(MPID_STATE_MPIR_COMM_RELEASE);

    MPID_MPI_FUNC_ENTER(MPID_STATE_MPIR_COMM_RELEASE);

    MPIR_Comm_release_ref( comm_ptr, &inuse );

    if (!inuse) {

	/* Remove the attributes, executing the attribute delete routine.  
           Do this only if the attribute functions are defined. 
	   This must be done first, because if freeing the attributes
	   returns an error, the communicator is not freed */
	if (MPIR_Process.attr_free && comm_ptr->attributes) {
	    /* Temporarily add a reference to this communicator because 
               the attr_free code requires a valid communicator */ 
	    MPIU_Object_add_ref( comm_ptr ); 
	    mpi_errno = MPIR_Process.attr_free( comm_ptr->handle, 
						comm_ptr->attributes );
	    /* Release the temporary reference added before the call to 
               attr_free */ 
	    MPIU_Object_release_ref( comm_ptr, &inuse); 
	}

	if (mpi_errno == MPI_SUCCESS) {
	    /* If this communicator is our parent, and we're disconnecting
	       from the parent, mark that fact */
	    if (MPIR_Process.comm_parent == comm_ptr)
		MPIR_Process.comm_parent = NULL;

	    /* Notify the device that the communicator is about to be 
	       destroyed */
	    MPID_Dev_comm_destroy_hook(comm_ptr);

#if !defined(FINEGRAIN_MPI)
	    /* Free the VCRT */
	    mpi_errno = MPID_VCRT_Release(comm_ptr->vcrt, isDisconnect);
	    if (mpi_errno != MPI_SUCCESS) {
		MPIU_ERR_POP(mpi_errno);
	    }
#endif
            if (comm_ptr->comm_kind == MPID_INTERCOMM) { /* HK: TODO INTERCOMM */
                mpi_errno = MPID_VCRT_Release(
		    comm_ptr->local_vcrt, isDisconnect);
		if (mpi_errno != MPI_SUCCESS) {
		    MPIU_ERR_POP(mpi_errno);
		}
                if (comm_ptr->local_comm) 
                    MPIR_Comm_release(comm_ptr->local_comm, isDisconnect );
            }

	    /* Free the context value */
	    MPIR_Free_contextid( comm_ptr->recvcontext_id ); 


	    /* Free the local and remote groups, if they exist */ 
            if (comm_ptr->local_group)
                MPIR_Group_release(comm_ptr->local_group);
            if (comm_ptr->remote_group)
                MPIR_Group_release(comm_ptr->remote_group);


  	    MPIU_Handle_obj_free( &MPID_Comm_mem, comm_ptr );  
	    
	    /* Remove from the list of active communicators if 
	       we are supporting message-queue debugging.  We make this
	       conditional on having debugger support since the
	       operation is not constant-time */
	    MPIR_COMML_FORGET( comm_ptr );
	}
	else {
	    /* If the user attribute free function returns an error,
	       then do not free the communicator */
	    MPIR_Comm_add_ref( comm_ptr );
	}
    }

 fn_exit:
    MPID_MPI_FUNC_EXIT(MPID_STATE_MPIR_COMM_RELEASE);
    return mpi_errno;
 fn_fail:
    goto fn_exit;
}
