/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*  $Id: init.c,v 1.34 2007/08/03 21:02:32 buntinas Exp $
 *
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

#include "mpiimpl.h"
#include "mpi_init.h"


/* -- Begin Profiling Symbol Block for routine MPI_Init */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Init = PMPI_Init
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Init  MPI_Init
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Init as PMPI_Init
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Init
#define MPI_Init PMPI_Init

/* Fortran logical values. extern'd in mpiimpl.h */
/* MPI_Fint MPIR_F_TRUE, MPIR_F_FALSE; */

/* Any internal routines can go here.  Make them static if possible */
#endif

#if defined(FINEGRAIN_MPI)
void* FG_Process_wrapper( void* wrapargs )
{
    FWraparg_t * wargs = (FWraparg_t*)wrapargs;
    int *fargc = &(wargs->argc);
    char*** fargv = &(wargs->argv);
    FG_ProcessPtr_t FG_MPI_processPtr = wargs->FG_FPtr;
    (*FG_MPI_processPtr)(*fargc, *fargv);
    return (0);
}


int FGmpiexec( int *argc, char ***argv, LookupMapPtr_t lookupFuncPtr )
{
    int mpi_errno = MPI_SUCCESS;
    int i, numfgps=-1, fgstartrank=-1, numspawn;
    char *mymapstr = NULL;
    FG_MapPtr_t FG_MAP_FunctionPtr = NULL;
    FG_ProcessPtr_t FG_MPI_processPtr = NULL;
    FG_WrapperProcessPtr_t WrapProcessPtr = NULL;
    FG_ProcessPtr_t FG_MPI_Main_processPtr = NULL;

    if (NULL == lookupFuncPtr){
        printf("A valid lookup function is missing.... bailing out\n");
        exit(-1);
    }

    mpi_errno = MPI_Init(argc, argv);

    MPIX_Get_collocated_size(&numfgps);
    assert(numfgps >= 1);
    numspawn = numfgps-1;
    MPIX_Get_collocated_startrank(&fgstartrank);
    assert(fgstartrank >= 0);

    mymapstr = getenv("FGMAP");
    FG_MAP_FunctionPtr = (FG_MapPtr_t)lookupFuncPtr(*argc, *argv, mymapstr);
    if (NULL == FG_MAP_FunctionPtr){
        printf("A valid map function is missing.... bailing out\n");
        exit(-1);
    }
    FG_MAP_FunctionPtr(*argc, *argv, MAP_INIT_ACTION);
    FG_MPI_Main_processPtr = FG_MAP_FunctionPtr(*argc, *argv, fgstartrank);

    if(numspawn > 0) {
        FG_Wrapargs = (FWraparg_t*)malloc(numspawn * sizeof(FWraparg_t));
        assert(FG_Wrapargs);

        /* mapping to functions */
        for (i=0; i< numspawn; i++) {
            FG_MPI_processPtr = FG_MAP_FunctionPtr(*argc, *argv, (i+1)+fgstartrank);
            FG_Wrapargs[i].argc = *argc;
            FG_Wrapargs[i].argv = *argv;
            FG_Wrapargs[i].FG_FPtr = FG_MPI_processPtr;
        }
        FG_MAP_FunctionPtr(*argc, *argv, MAP_FINALIZE_ACTION);

        WrapProcessPtr = &FG_Process_wrapper;
        FG_Main_Init(WrapProcessPtr, FG_Wrapargs, numspawn, argc, argv);
    }

    (*FG_MPI_Main_processPtr)(*argc, *argv);

    return (mpi_errno);
}
#endif

#undef FUNCNAME
#define FUNCNAME MPI_Init

/*@
   MPI_Init - Initialize the MPI execution environment

   Input Parameters:
+  argc - Pointer to the number of arguments 
-  argv - Pointer to the argument vector

Thread and Signal Safety:
This routine must be called by one thread only.  That thread is called
the `main thread` and must be the thread that calls 'MPI_Finalize'.

Notes:
   The MPI standard does not say what a program can do before an 'MPI_INIT' or
   after an 'MPI_FINALIZE'.  In the MPICH implementation, you should do
   as little as possible.  In particular, avoid anything that changes the
   external state of the program, such as opening files, reading standard
   input or writing to standard output.

Notes for Fortran:
The Fortran binding for 'MPI_Init' has only the error return
.vb
    subroutine MPI_INIT( ierr )
    integer ierr
.ve

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_INIT

.seealso: MPI_Init_thread, MPI_Finalize
@*/

int MPI_Init( int *argc, char ***argv )
{
    static const char FCNAME[] = "MPI_Init";
    int mpi_errno = MPI_SUCCESS;
    MPID_MPI_INIT_STATE_DECL(MPID_STATE_MPI_INIT);

    MPID_CS_INITIALIZE();
    /* FIXME: Can we get away without locking every time.  Now, we
       need a MPID_CS_ENTER/EXIT around MPI_Init and MPI_Init_thread.
       Progress may be called within MPI_Init, e.g., by a spawned
       child process.  Within progress, the lock is released and
       reacquired when blocking.  If the lock isn't acquired before
       then, the release in progress is incorrect.  Furthermore, if we
       don't release the lock after progress, we'll deadlock the next
       time this process tries to acquire the lock.
       MPID_CS_ENTER/EXIT functions are used here instead of
       MPIU_THREAD_SINGLE_CS_ENTER/EXIT because
       MPIR_ThreadInfo.isThreaded hasn't been initialized yet.
    */
    MPID_CS_ENTER();
    
    MPID_MPI_INIT_FUNC_ENTER(MPID_STATE_MPI_INIT);

#if defined(FINEGRAIN_MPI)

       /* HK: Initializing of FGP StateWrapper takes place here.
          The following if-statement makes sure that allocation to statevars per FGP is only done once,
          hence an FGP will not be allowed to call MPI_Init() more than once.
       */    
       if( 0 == CO_CURRENT->statevars){
           /* MPIU_Calloc allocates memory that is initialized to zero, so MPIR_ThreadFG will be initialized to zero.*/
           CO_CURRENT->statevars = (struct StateWrapper*)MPIU_Calloc(1, sizeof(struct StateWrapper));
           MPIU_Assert (CO_CURRENT->statevars != 0);
       }

       if(0 == INIT_INITIALIZED){
           MPIR_Process.initialized = MPICH_PRE_INIT;
           INIT_INITIALIZED = 1;           
       }
       
       if (FGP_PRE_INIT == FGP_init_state){           
           FGP_init_state = FGP_WITHIN_INIT; /* HK: The first FGP sets the initialization state. */
           IS_SPAWNER = 1;
       }

#endif

#   ifdef HAVE_ERROR_CHECKING
    {
        MPID_BEGIN_ERROR_CHECKS;
        {
            if (MPIR_Process.initialized != MPICH_PRE_INIT) {
#if defined(FINEGRAIN_MPI)
                if(1 == IS_SPAWNER){
                    assert( (FGP_POST_INIT == FGP_init_state) ||
                            (FGP_ALL_POST_INIT == FGP_init_state) );
                    IS_SPAWNER = 2; /* IS_SPAWNER equal to 2 means this is the
                                       expected (silent) MPI_Init()and it has now been seen. */
                    return (mpi_errno);
                } else if ((2 == IS_SPAWNER) || (0 == IS_SPAWNER)){
                    printf("WARNING! Application rank=%d is calling MPI_Init() again after initialization!\n", my_fgrank);                    
                    goto fn_fail;
                } else {
                    printf("%d: ERROR! This code should not be reached! Exiting....\n", my_fgrank);
                    exit(-1);
                }
#else
                mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER,
						  "**inittwice", NULL );
#endif
	    }
            if (mpi_errno) goto fn_fail;
        }
        MPID_END_ERROR_CHECKS;
    }
#   endif /* HAVE_ERROR_CHECKING */

    /* ... body of routine ... */

#if defined(FINEGRAIN_MPI)
       mask_in_use = 0;
       lowestContextId = MPIR_MAXID;
       initialize_context_mask = 1;
       /* HK: context_mask array and local_mask array in MPIR_Get_contextid() have to be declared
          dynamically in order to support a large number of context-ids. The previously used static array
          can be only of a limited size. For example, a static array of size 32768 fails on this setup.
        */
       context_mask = NULL; /* HK: This is allocated in MPIR_Init_contextid */

       LBI_mask = NULL; /* HK: This is allocated in MPIR_Init_LBI */
       initialize_LBI_mask = 1;
#endif

    mpi_errno = MPIR_Init_thread( argc, argv, MPI_THREAD_SINGLE, (int *)0 );
    if (mpi_errno != MPI_SUCCESS) goto fn_fail;

    /* ... end of body of routine ... */

    MPID_MPI_INIT_FUNC_EXIT(MPID_STATE_MPI_INIT);
    MPID_CS_EXIT();

#if defined(FINEGRAIN_MPI)
    if(FGP_WITHIN_INIT == FGP_init_state)
    {
        FGP_init_state = FGP_POST_INIT;
    } else {
        assert(FGP_POST_INIT == FGP_init_state);
        FG_Init();
    }

#endif

    return mpi_errno;
    
  fn_fail:
    /* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_REPORTING
    {
	mpi_errno = MPIR_Err_create_code(
	    mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
	    "**mpi_init", "**mpi_init %p %p", argc, argv);
    }
#   endif
    mpi_errno = MPIR_Err_return_comm( 0, FCNAME, mpi_errno );
    MPID_CS_EXIT();
    MPID_CS_FINALIZE();
    return mpi_errno;
    /* --END ERROR HANDLING-- */
}
